"""
disk device support class(es)

http://libvirt.org/formatdomain.html#elementsDisks
"""

from virttest.libvirt_xml import accessors, xcepts
from virttest.libvirt_xml.devices import base, librarian


class Disk(base.TypedDeviceBase):

    """
    Disk device XML class

    Properties:
        device:
            string, how exposted to guest
        rawio:
            string (yes/no), disk needs rawio capability
        sgio:
            string, "filtered" or "unfiltered"
        snapshot:
            string, "yes", "no", "internal" or "external"
        wwn:
            string.
        serial:
            string.
        vendor:
            string.
        product:
            string.
        driver:
            dict, keys: name, type, cache, error_policy, io, ioeventfd,
            event_idx, copy_on_read, discard
        target:
            dict, keys: dev, bus, tray
        alias:
            dict, keys: name
        blockio:
            dict, keys: logical_block_size, physical_block_size
        geometry:
            dict, keys: cyls, heads, secs, trans
        address:
            libvirt_xml.devices.Address instance
        boot:
            string, boot order number to use if not using boot in os element
        loadparm:
            string, loadparm attribute on disk's boot element
        readonly:
            bool, True/False
        transient:
            bool, True/False
        share:
            bool, True/False
        mirror:
            bool, read-only, True if block copy started
        ready:
            bool, read-only, True if disk ready for pivot
        iotune:
            libvirt_xml.devices.Disk.IOTune instance
        source:
            libvirt_xml.devices.Disk.DiskSource instance
        encryption:
            libvirt_xml.devices.Disk.Encryption instance.
        auth:
            libvirt_xml.devices.Disk.Auth instance.
        reservations:
            libvirt_xml.devices.Disk.Reservations instance.
   """

    __slots__ = ('device', 'rawio', 'sgio', 'snapshot', 'driver', 'target', 'alias',
                 'address', 'boot', 'loadparm', 'readonly', 'transient', 'share', 'model',
                 'mirror', 'ready', 'iotune', 'source', 'blockio', 'geometry',
                 'wwn', 'serial', 'vendor', 'product', 'encryption', 'auth',
                 'reservations', 'backingstore')

    def __init__(self, type_name='file', virsh_instance=base.base.virsh):
        accessors.XMLAttribute('device', self, parent_xpath='/',
                               tag_name='disk', attribute='device')
        accessors.XMLAttribute('model', self, parent_xpath='/',
                               tag_name='disk', attribute='model')
        accessors.XMLAttribute('rawio', self, parent_xpath='/',
                               tag_name='disk', attribute='rawio')
        accessors.XMLAttribute('sgio', self, parent_xpath='/',
                               tag_name='disk', attribute='sgio')
        accessors.XMLAttribute('snapshot', self, parent_xpath='/',
                               tag_name='disk', attribute='snapshot')
        accessors.XMLElementText('wwn', self, parent_xpath='/',
                                 tag_name='wwn')
        accessors.XMLElementText('serial', self, parent_xpath='/',
                                 tag_name='serial')
        accessors.XMLElementText('vendor', self, parent_xpath='/',
                                 tag_name='vendor')
        accessors.XMLElementText('product', self, parent_xpath='/',
                                 tag_name='product')
        accessors.XMLElementDict('driver', self, parent_xpath='/',
                                 tag_name='driver')
        accessors.XMLElementDict('target', self, parent_xpath='/',
                                 tag_name='target')
        accessors.XMLElementDict('alias', self, parent_xpath='/',
                                 tag_name='alias')
        accessors.XMLElementDict('blockio', self, parent_xpath='/',
                                 tag_name='blockio')
        accessors.XMLElementDict('geometry', self, parent_xpath='/',
                                 tag_name='geometry')
        accessors.XMLElementNest('address', self, parent_xpath='/',
                                 tag_name='address', subclass=self.Address,
                                 subclass_dargs={'type_name': 'drive',
                                                 'virsh_instance': virsh_instance})
        accessors.XMLAttribute('boot', self, parent_xpath='/',
                               tag_name='boot', attribute='order')
        accessors.XMLAttribute('loadparm', self, parent_xpath='/',
                               tag_name='boot', attribute='loadparm')
        accessors.XMLElementBool('readonly', self, parent_xpath='/',
                                 tag_name='readonly')
        accessors.XMLElementBool('transient', self, parent_xpath='/',
                                 tag_name='transient')
        accessors.XMLElementBool('share', self, parent_xpath='/',
                                 tag_name='shareable')
        accessors.XMLElementNest('source', self, parent_xpath='/',
                                 tag_name='source', subclass=self.DiskSource,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        ro = ['set', 'del']
        accessors.XMLElementBool('mirror', self, forbidden=ro,
                                 parent_xpath='/', tag_name='mirror')
        accessors.XMLElementBool('ready', self, forbidden=ro,
                                 parent_xpath='/', tag_name='ready')
        accessors.XMLElementNest('iotune', self, parent_xpath='/',
                                 tag_name='iotune', subclass=self.IOTune,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        accessors.XMLElementNest('encryption', self, parent_xpath='/',
                                 tag_name='encryption', subclass=self.Encryption,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        accessors.XMLElementNest('auth', self, parent_xpath='/',
                                 tag_name='auth', subclass=self.Auth,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        accessors.XMLElementNest('reservations', self, parent_xpath='/',
                                 tag_name='reservations',
                                 subclass=Disk.Reservations,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        accessors.XMLElementNest('backingstore', self, parent_xpath='/',
                                 tag_name='backingStore',
                                 subclass=self.BackingStore,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        super(Disk, self).__init__(device_tag='disk', type_name=type_name,
                                   virsh_instance=virsh_instance)

    def new_disk_source(self, **dargs):
        """
        Return a new disk source instance and set properties from dargs
        """
        new_one = self.DiskSource(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def new_iotune(self, **dargs):
        """
        Return a new disk IOTune instance and set properties from dargs
        """
        new_one = self.IOTune(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def new_encryption(self, **dargs):
        """
        Return a new disk encryption instance and set properties from dargs
        """
        new_one = self.Encryption(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def new_disk_address(self, type_name='drive', **dargs):
        """
        Return a new disk Address instance and set properties from dargs
        """
        new_one = self.Address(type_name=type_name, virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def new_auth(self, **dargs):
        """
        Return a new disk auth instance and set properties from dargs
        """
        new_one = self.Auth(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def new_reservations(self, **dargs):
        """
        Return a new disk reservations instance and set properties from dargs
        """
        new_one = self.Reservations(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def new_slices(self, **dargs):
        """
        Return a new disk slices instance and set properties from dargs
        """
        new_one = self.Slices(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def new_backingstore(self, **dargs):
        """
        Return a new disk backingstore instance and set properties from dargs
        """
        new_one = self.BackingStore(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    # For convenience
    Address = librarian.get('address')

    class DiskSource(base.base.LibvirtXMLBase):

        """
        Disk source device XML class

        Properties:

        attrs: Dictionary of attributes, qualifying the disk type
        seclabels: list of libvirt_xml.devices.seclabel.Seclabel instances
        hosts: list of dictionaries describing network host properties
        """

        __slots__ = ('attrs', 'seclabels', 'hosts', 'encryption', 'auth',
                     'reservations', 'slices', 'config_file', 'snapshot_name')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLElementDict('attrs', self, parent_xpath='/',
                                     tag_name='source')
            accessors.XMLElementList('seclabels', self, parent_xpath='/',
                                     marshal_from=self.marshal_from_seclabel,
                                     marshal_to=self.marshal_to_seclabel)
            accessors.XMLElementList('hosts', self, parent_xpath='/',
                                     marshal_from=self.marshal_from_host,
                                     marshal_to=self.marshal_to_host)
            accessors.XMLElementNest('encryption', self, parent_xpath='/',
                                     tag_name='encryption',
                                     subclass=Disk.Encryption,
                                     subclass_dargs={
                                         'virsh_instance': virsh_instance})
            accessors.XMLElementNest('auth', self, parent_xpath='/',
                                     tag_name='auth', subclass=Disk.Auth,
                                     subclass_dargs={
                                         'virsh_instance': virsh_instance})
            accessors.XMLElementNest('reservations', self, parent_xpath='/',
                                     tag_name='reservations',
                                     subclass=Disk.Reservations,
                                     subclass_dargs={
                                         'virsh_instance': virsh_instance})
            accessors.XMLElementNest('slices', self, parent_xpath='/',
                                     tag_name='slices',
                                     subclass=Disk.Slices,
                                     subclass_dargs={
                                         'virsh_instance': virsh_instance})
            accessors.XMLAttribute('config_file', self, parent_xpath='/',
                                   tag_name='config', attribute='file')
            accessors.XMLAttribute('snapshot_name', self, parent_xpath='/',
                                   tag_name='snapshot', attribute='name')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<source/>'

        @staticmethod
        def marshal_from_seclabel(item, index, libvirtxml):
            """Convert a Seclabel instance into tag + attributes"""
            del index           # not used
            del libvirtxml      # not used
            root = item.xmltreefile.getroot()
            if root.tag == 'seclabel':
                new_dict = dict(list(root.items()))
                text_dict = {}
                # Put element text into dict under key 'text'
                for key in ('label', 'baselabel'):
                    text_val = item.xmltreefile.findtext(key)
                    if text_val:
                        text_dict.update({key: text_val})
                if text_dict:
                    new_dict['text'] = text_dict
                return (root.tag, new_dict)
            else:
                raise xcepts.LibvirtXMLError("Expected a list of seclabel "
                                             "instances, not a %s" % str(item))

        @staticmethod
        def marshal_to_seclabel(tag, attr_dict, index, libvirtxml):
            """Convert a tag + attributes into a Seclabel instance"""
            del index           # not used
            if tag != 'seclabel':
                return None     # Don't convert this item
            Seclabel = librarian.get('seclabel')
            newone = Seclabel(virsh_instance=libvirtxml.virsh)
            newone.update(attr_dict)
            return newone

        @staticmethod
        def marshal_from_host(item, index, libvirtxml):
            """Convert a dictionary into a tag + attributes"""
            del index           # not used
            del libvirtxml      # not used
            if not isinstance(item, dict):
                raise xcepts.LibvirtXMLError("Expected a dictionary of host "
                                             "attributes, not a %s"
                                             % str(item))
            return ('host', dict(item))  # return copy of dict, not reference

        @staticmethod
        def marshal_to_host(tag, attr_dict, index, libvirtxml):
            """Convert a tag + attributes into a dictionary"""
            del index                    # not used
            del libvirtxml               # not used
            if tag != 'host':
                return None              # skip this one
            return dict(attr_dict)       # return copy of dict, not reference

    class IOTune(base.base.LibvirtXMLBase):

        """
        IOTune device XML class

        Properties:

        total_bytes_sec: str(int)
        read_bytes_sec: str(int)
        write_bytes_sec: str(int)
        total_iops_sec: str(int)
        read_iops_sec: str(int)
        write_iops_sec: str(int)
        """

        __slots__ = ('total_bytes_sec', 'read_bytes_sec', 'write_bytes_sec',
                     'total_iops_sec', 'read_iops_sec', 'write_iops_sec')

        def __init__(self, virsh_instance=base.base.virsh):
            for slot in self.__all_slots__:
                if slot in base.base.LibvirtXMLBase.__all_slots__:
                    continue    # don't add these
                else:
                    accessors.XMLElementInt(slot, self, parent_xpath='/',
                                            tag_name=slot)
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<iotune/>'

    class Encryption(base.base.LibvirtXMLBase):

        """
        Encryption device XML class

        Properties:

        encryption:
            string.
        secret:
            dict, keys: type, uuid
        """

        __slots__ = ('encryption', 'secret')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLAttribute('encryption', self, parent_xpath='/',
                                   tag_name='encryption', attribute='format')
            accessors.XMLElementDict('secret', self, parent_xpath='/',
                                     tag_name='secret')
            super(self.__class__, self).__init__(
                virsh_instance=virsh_instance)
            self.xml = '<encryption/>'

    class Auth(base.base.LibvirtXMLBase):

        """
        Auth device XML class

        Properties:

        auth_user:
            string, attribute of auth tag
        secret_type:
            string, attribute of secret tag, sub-tag of the auth tag
        secret_uuid:
            string, attribute of secret tag, sub-tag of the auth tag
        secret_usage:
            string, attribute of secret tag, sub-tag of the auth tag
        """

        __slots__ = ('auth_user', 'secret_type', 'secret_uuid', 'secret_usage')

        def __init__(self, virsh_instance=base.base.virsh, auth_user=""):
            accessors.XMLAttribute('auth_user', self, parent_xpath='/',
                                   tag_name='auth', attribute='username')
            accessors.XMLAttribute('secret_type', self, parent_xpath='/',
                                   tag_name='secret', attribute='type')
            accessors.XMLAttribute('secret_uuid', self, parent_xpath='/',
                                   tag_name='secret', attribute='uuid')
            accessors.XMLAttribute('secret_usage', self, parent_xpath='/',
                                   tag_name='secret', attribute='usage')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<auth/>'

    class Slices(base.base.LibvirtXMLBase):

        """
        slices device XML class
        Typical xml looks like:
        <slices>
          <slice type='storage' offset='12345' size='123'/>
        </slices>
        Properties:

        slice_type:
            string, type attribute of slice tag
        slice_offset:
            string, offset attribute of slice tag
        slice_size:
            string, size attribute of slice tag
        """

        __slots__ = ('slice_type', 'slice_offset', 'slice_size')

        def __init__(self, virsh_instance=base.base.virsh, auth_user=""):
            accessors.XMLAttribute('slice_type', self, parent_xpath='/',
                                   tag_name='slice', attribute='type')
            accessors.XMLAttribute('slice_offset', self, parent_xpath='/',
                                   tag_name='slice', attribute='offset')
            accessors.XMLAttribute('slice_size', self, parent_xpath='/',
                                   tag_name='slice', attribute='size')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<slices/>'

    class Reservations(base.base.LibvirtXMLBase):

        """
        Reservations device XML class

        Properties:

        reservations_managed:
            string, attribute of reservations tag
        reservations_source_type:
            string, attribute of source tag, sub-tag of the reservations tag
        reservations_source_path:
            string, attribute of source tag, sub-tag of the reservations tag
        reservations_source_mode:
            string, attribute of source tag, sub-tag of the reservations tag
        """

        __slots__ = ('reservations_managed', 'reservations_source_type',
                     'reservations_source_path', 'reservations_source_mode')

        def __init__(self, virsh_instance=base.base.virsh, reservations_managed=""):
            accessors.XMLAttribute('reservations_managed', self, parent_xpath='/',
                                   tag_name='reservations', attribute='managed')
            accessors.XMLAttribute('reservations_source_type', self, parent_xpath='/',
                                   tag_name='source', attribute='type')
            accessors.XMLAttribute('reservations_source_path', self, parent_xpath='/',
                                   tag_name='source', attribute='path')
            accessors.XMLAttribute('reservations_source_mode', self, parent_xpath='/',
                                   tag_name='source', attribute='mode')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<reservations/>'

    class BackingStore(base.base.LibvirtXMLBase):
        """
        BakingStore of disk device XML class

        type:
            string, attribute of backingStore tag
        index:
            string, attribute of backingStore tag
        format:
            dict, key-attribute of backingStore tag
        source:
            nested xml of backingStore tag
        """
        __slots__ = ('type', 'index', 'format', 'source')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLAttribute('type', self,
                                   parent_xpath='/',
                                   tag_name='backingStore',
                                   attribute='type')
            accessors.XMLAttribute('index', self,
                                   parent_xpath='/',
                                   tag_name='backingStore',
                                   attribute='index')
            accessors.XMLElementDict('format', self,
                                     parent_xpath='/',
                                     tag_name='format')
            accessors.XMLElementNest('source', self,
                                     parent_xpath='/',
                                     tag_name='source',
                                     subclass=self.Source,
                                     subclass_dargs={
                                         'virsh_instance': virsh_instance})

            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<backingStore/>'

        def new_source(self, **dargs):
            """
            Create new source for backingstore

            """
            new_one = self.Source(virsh_instance=self.virsh)
            for key, value in list(dargs.items()):
                setattr(new_one, key, value)
            return new_one

        class Source(base.base.LibvirtXMLBase):
            """
            Source of backingstore xml class

            dev:
                string, attribute of backingStore/source tag
            protocal:
                string, attribute of backingStore/source tag
            name:
                string, attribute of backingStore/source tag
            host:
                dict, nested xml of backingStore/source tag
            file:
                string, attribute of backingStore/source tag
            """

            __slots__ = ('dev', 'protocol', 'name', 'host', 'file', 'auth')

            def __init__(self, virsh_instance=base.base.virsh):
                accessors.XMLAttribute('dev', self,
                                       parent_xpath='/',
                                       tag_name='source',
                                       attribute='dev')
                accessors.XMLAttribute('protocol', self,
                                       parent_xpath='/',
                                       tag_name='source',
                                       attribute='protocol')
                accessors.XMLAttribute('name', self,
                                       parent_xpath='/',
                                       tag_name='source',
                                       attribute='name')
                accessors.XMLElementDict('host', self,
                                         parent_xpath='/',
                                         tag_name='host')
                accessors.XMLAttribute('file', self,
                                       parent_xpath='/',
                                       tag_name='source',
                                       attribute='file')
                accessors.XMLElementNest('auth', self,
                                         parent_xpath='/',
                                         tag_name='auth',
                                         subclass=Disk.Auth,
                                         subclass_dargs={
                                            'virsh_instance': virsh_instance})

                super(self.__class__, self).__init__(virsh_instance=virsh_instance)
                self.xml = '<source/>'



"""
Module simplifying manipulation of XML described at
http://libvirt.org/formatcaps.html
"""

from virttest import xml_utils
from virttest.libvirt_xml import base, accessors, xcepts


class CapabilityXML(base.LibvirtXMLBase):

    """
    Handler of libvirt capabilities and nonspecific item operations.

    Properties:
        uuid:
            string of host uuid
        guest_capabilities:
            dict, read-only
        get:
            dict map from os type names to dict map from arch names
    """

    # TODO: Add more __slots__ and accessors to get some useful stats
    # e.g. guest_count etc.

    __slots__ = ('uuid', 'guest_capabilities', 'cpu_count', 'arch', 'model',
                 'vendor', 'feature_list', 'power_management_list',
                 'cpu_topology', 'cells_topology')
    __schema_name__ = "capability"

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLElementText(property_name="uuid",
                                 libvirtxml=self,
                                 forbidden=['set', 'del'],
                                 parent_xpath='/host',
                                 tag_name='uuid')
        # This will skip self.get_guest_capabilities() defined below
        accessors.AllForbidden(property_name="guest_capabilities",
                               libvirtxml=self)
        # This will skip self.get_cpu_count() defined below
        accessors.AllForbidden(property_name="cpu_count",
                               libvirtxml=self)
        # The set action is for test.
        accessors.XMLElementText(property_name="arch",
                                 libvirtxml=self,
                                 forbidden=['del'],
                                 parent_xpath='/host/cpu',
                                 tag_name='arch')
        accessors.XMLElementText(property_name="model",
                                 libvirtxml=self,
                                 forbidden=['del'],
                                 parent_xpath='/host/cpu',
                                 tag_name='model')
        accessors.XMLElementText(property_name="vendor",
                                 libvirtxml=self,
                                 forbidden=['del'],
                                 parent_xpath='/host/cpu',
                                 tag_name='vendor')
        accessors.XMLElementDict(property_name="cpu_topology",
                                 libvirtxml=self,
                                 forbidden=['del'],
                                 parent_xpath='/host/cpu',
                                 tag_name='topology')
        accessors.XMLElementNest(property_name='cells_topology',
                                 libvirtxml=self,
                                 parent_xpath='/host',
                                 tag_name='topology',
                                 subclass=TopologyXML,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        # This will skip self.get_feature_list() defined below
        accessors.AllForbidden(property_name="feature_list",
                               libvirtxml=self)
        # This will skip self.get_power_management_list() defined below
        accessors.AllForbidden(property_name="power_management_list",
                               libvirtxml=self)
        super(CapabilityXML, self).__init__(virsh_instance)
        # calls set_xml accessor method
        self['xml'] = self.__dict_get__('virsh').capabilities()

    def get_guest_capabilities(self):
        """
        Accessor method for guest_capabilities property (in __slots__).
        Return a guest capabilities dict in following schema:
        {<os_type>: {<arch name>: {'wordsize': '', 'emulator': '',
        'machine': [<machine name>, ...], 'domaini_<type>': {'emulator': ''}}}}
        """
        guest_capa = {}
        xmltreefile = self.__dict_get__('xml')
        for guest in xmltreefile.findall('guest'):
            os_type_name = guest.find('os_type').text
            # Multiple guest definitions can share same os_type (e.g. hvm, pvm)
            if os_type_name == 'xen':
                os_type_name = 'pv'
            guest_arch = guest_capa.get(os_type_name, {})
            for arch in guest.findall('arch'):
                arch_name = arch.get('name')
                arch_prop = guest_arch.get(arch_name, {})
                arch_prop['wordsize'] = arch.find('wordsize').text
                arch_prop['emulator'] = arch.find('emulator').text
                # TODO: maxcpus differs for machine in a given arch.
                # Right now takes first machine as default for the given
                # arch and assigns maxcpu for the arch, ideally below machine
                # list to get popluated with maxcpu for each machine
                # not modifying as no requirement for now and as it would break
                # other places where machine list is used.
                arch_prop['maxcpus'] = arch.findall('machine')[0].get('maxCpus')
                m_list = []
                for machine in arch.findall('machine'):
                    machine_text = machine.text
                    # Don't add duplicate entries
                    if not m_list.count(machine_text):
                        m_list.append(machine_text)
                arch_prop['machine'] = m_list
                for domain in arch.findall('domain'):
                    domain_name = "domain_" + domain.get('type')
                    dom_prop = {}
                    if domain.find('emulator') is not None:
                        dom_prop['emulator'] = domain.find('emulator').text
                    arch_prop[domain_name] = dom_prop
                guest_arch[arch_name] = arch_prop
            guest_capa[os_type_name] = guest_arch
        return guest_capa

    def get_power_management_list(self):
        """
        Accessor method for power_management_list property (in __slots__)
        """
        xmltreefile = self.__dict_get__('xml')
        pms = xmltreefile.find('host').find('power_management').getchildren()
        return [pm.tag for pm in pms]

    def get_feature_list(self):
        """
        Accessor method for feature_list property (in __slots__)
        """
        feature_list = []  # [<feature1>, <feature2>, ...]
        xmltreefile = self.__dict_get__('xml')
        for feature_node in xmltreefile.findall('/host/cpu/feature'):
            feature_list.append(feature_node)
        return feature_list

    def get_feature(self, num):
        """
        Get a feature element from feature list by number

        :return: Feature element
        """
        count = len(self.feature_list)
        try:
            num = int(num)
            return self.feature_list[num]
        except (ValueError, TypeError):
            raise xcepts.LibvirtXMLError("Invalid feature number %s" % num)
        except IndexError:
            raise xcepts.LibvirtXMLError("Only %d feature(s)" % count)

    def get_feature_name(self, num):
        """
        Get assigned feature name

        :param num: Assigned feature number
        :return: Assigned feature name
        """
        return self.get_feature(num).get('name')

    def get_cpu_count(self):
        """
        Accessor method for cpu_count property (in __slots__)
        """
        cpu_count = 0
        xmltreefile = self.__dict_get__('xml')
        for cpus in xmltreefile.findall('/host/topology/cells/cell/cpus'):
            cpu_num = cpus.get('num')
            cpu_count += int(cpu_num)
        return cpu_count

    def remove_feature(self, num):
        """
        Remove a assigned feature from xml

        :param num: Assigned feature number
        """
        xmltreefile = self.__dict_get__('xml')
        feature_remove_node = self.get_feature(num)
        cpu_node = xmltreefile.find('/host/cpu')
        cpu_node.remove(feature_remove_node)

    def check_feature_name(self, name):
        """
        Check feature name valid or not.

        :param name: The checked feature name
        :return: True if check pass
        """
        sys_feature = []
        cpu_xml_file = open('/proc/cpuinfo', 'r')
        for line in cpu_xml_file.readlines():
            if line.find('flags') != -1:
                feature_names = line.split(':')[1].strip()
                sys_sub_feature = feature_names.split(' ')
                sys_feature = list(set(sys_feature + sys_sub_feature))
        cpu_xml_file.close()
        return (name in sys_feature)

    def set_feature(self, num, value):
        """
        Set a assigned feature value to xml

        :param num: Assigned feature number
        :param value: The feature name modified to
        """
        feature_set_node = self.get_feature(num)
        feature_set_node.set('name', value)

    def add_feature(self, value):
        """
        Add a feature Element to xml

        :param value: The added feature name
        """
        xmltreefile = self.__dict_get__('xml')
        cpu_node = xmltreefile.find('/host/cpu')
        xml_utils.ElementTree.SubElement(cpu_node, 'feature', {'name': value})


class TopologyXML(base.LibvirtXMLBase):

    """
    Handler of cells topology element in libvirt capabilities.

    Properties:
        num:
            string of node cell numbers
        cell:
            list of cpu dict
    """

    __slots__ = ('num', 'cell')

    def __init__(self, virsh_instance=base.virsh):
        """
        Create new cells topology XML instance
        """
        accessors.XMLAttribute(property_name="num",
                               libvirtxml=self,
                               parent_xpath='/',
                               tag_name='cells',
                               attribute='num')
        accessors.AllForbidden(property_name="cell",
                               libvirtxml=self)
        super(TopologyXML, self).__init__(virsh_instance)
        self.xml = self.__dict_get__('virsh').capabilities()
        self.xmltreefile.reroot("/host/topology")
        self.xmltreefile.write()

    def get_cell(self, withmem=False):
        """
        Return CellXML instances list
        """
        cell_list = []
        for cell_node in self.xmltreefile.findall('/cells/cell'):
            if not cell_node.find('memory') and withmem:
                continue
            xml_str = xml_utils.ElementTree.tostring(
                cell_node)
            new_cell = CellXML()
            new_cell.xml = xml_str
            cell_list.append(new_cell)
        return cell_list


class CellXML(base.LibvirtXMLBase):

    """
    Handler of cell element in libvirt capabilities.

    Properties:
        cell_id:
            string of node cell number id
        memory:
            int, memory size
        mem_unit:
            string of memory unit
        pages:
            list of pages dict
        sibling:
            list of sibling dict
        cpus_num:
            string of cpus number
        cpu:
            list of cpu dict
    """

    __slots__ = ('cell_id', 'memory', 'mem_unit', 'pages', 'sibling',
                 'cpus_num', 'cpu')

    def __init__(self, virsh_instance=base.virsh):
        """
        Create new cpus XML instance
        """
        accessors.XMLAttribute(property_name="cell_id",
                               libvirtxml=self,
                               parent_xpath='/',
                               tag_name='cell',
                               attribute='id')
        accessors.XMLElementInt(property_name="memory",
                                libvirtxml=self,
                                parent_xpath='/',
                                tag_name='memory')
        accessors.XMLAttribute(property_name="mem_unit",
                               libvirtxml=self,
                               parent_xpath='/',
                               tag_name='memory',
                               attribute='unit')
        accessors.XMLElementList(property_name="pages",
                                 libvirtxml=self,
                                 parent_xpath='/',
                                 marshal_from=self.marshal_from_pages,
                                 marshal_to=self.marshal_to_pages)
        accessors.XMLElementList(property_name="sibling",
                                 libvirtxml=self,
                                 parent_xpath='/distances',
                                 marshal_from=self.marshal_from_sibling,
                                 marshal_to=self.marshal_to_sibling)
        accessors.XMLAttribute(property_name="cpus_num",
                               libvirtxml=self,
                               parent_xpath='/',
                               tag_name='cpus',
                               attribute='num')
        accessors.XMLElementList(property_name="cpu",
                                 libvirtxml=self,
                                 parent_xpath='/cpus',
                                 marshal_from=self.marshal_from_cpu,
                                 marshal_to=self.marshal_to_cpu)
        super(CellXML, self).__init__(virsh_instance)
        self.xml = u"<cell></cell>"

    @staticmethod
    def marshal_from_pages(item, index, libvirtxml):
        """
        Convert a dict to pages tag and attributes.
        """
        del index
        del libvirtxml
        if not isinstance(item, dict):
            raise xcepts.LibvirtXMLError("Expected a dictionary of pages "
                                         "attributes, not a %s"
                                         % str(item))
        return ('pages', dict(item))

    @staticmethod
    def marshal_to_pages(tag, attr_dict, index, libvirtxml, text):
        """
        Convert a pages tag and attributes to a dict.
        """
        del index
        del libvirtxml
        if tag != 'pages':
            return None
        attr_dict['text'] = text
        return dict(attr_dict)

    @staticmethod
    def marshal_from_sibling(item, index, libvirtxml):
        """
        Convert a dict to sibling tag and attributes.
        """
        del index
        del libvirtxml
        if not isinstance(item, dict):
            raise xcepts.LibvirtXMLError("Expected a dictionary of sibling "
                                         "attributes, not a %s"
                                         % str(item))
        return ('sibling', dict(item))

    @staticmethod
    def marshal_to_sibling(tag, attr_dict, index, libvirtxml):
        """
        Convert a sibling tag and attributes to a dict.
        """
        del index
        del libvirtxml
        if tag != 'sibling':
            return None
        return dict(attr_dict)

    @staticmethod
    def marshal_from_cpu(item, index, libvirtxml):
        """
        Convert a dict to cpu tag and attributes.
        """
        del index
        del libvirtxml
        if not isinstance(item, dict):
            raise xcepts.LibvirtXMLError("Expected a dictionary of cpu "
                                         "attributes, not a %s"
                                         % str(item))
        return ('cpu', dict(item))

    @staticmethod
    def marshal_to_cpu(tag, attr_dict, index, libvirtxml):
        """
        Convert a cpu tag and attributes to a dict.
        """
        del index
        del libvirtxml
        if tag != 'cpu':
            return None
        return dict(attr_dict)



"""
arp protocl support class(es)

http://libvirt.org/formatnwfilter.html#nwfelemsRulesProtoARP
"""

from virttest.libvirt_xml import accessors, xcepts
from virttest.libvirt_xml.nwfilter_protocols import base


class Arp(base.TypedDeviceBase):

    """
    Create new Arp xml instances

    Properties:

    attrs: libvirt_xml.nwfilter_protocols.Arp.Attr instance
    """

    __slots__ = ('attrs',)

    def __init__(self, type_name='file', virsh_instance=base.base.virsh):
        accessors.XMLElementNest('attrs', self, parent_xpath='/',
                                 tag_name='arp', subclass=self.Attr,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        super(Arp, self).__init__(protocol_tag='arp', type_name=type_name,
                                  virsh_instance=virsh_instance)

    def new_attr(self, **dargs):
        """
        Return a new Attr instance and set properties from dargs

        :param dargs: dict of attributes
        :return: new Attr instance
        """
        new_one = self.Attr(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def get_attr(self):
        """
        Return arp attribute dict

        :return: None if no arp in xml, dict of arp's attributes.
        """
        try:
            arp_node = self.xmltreefile.reroot('/arp')
        except KeyError as detail:
            raise xcepts.LibvirtXMLError(detail)
        node = arp_node.getroot()
        arp_attr = dict(list(node.items()))

        return arp_attr

    class Attr(base.base.LibvirtXMLBase):

        """
        Arp attribute XML class

        Properties:

        srcmacaddr: string, MAC address of sender
        srcmacmask: string, Mask applied to MAC address of sender
        dstmacaddr: string, MAC address of destination
        dstmacaddr: string, Mask applied to MAC address of destination
        hwtype: string, Hardware type
        protocoltype: string, Protocol type
        opcode: string, Opcode
        arpsrcmacaddr: string, Source MAC address in ARP/RARP packet
        arpdstmacaddr: string, Destination MAC address in ARP/RARP packet
        arpsrcipaddr: string, Source IP address in ARP/RARP packet
        arpdstipaddr: string, Destination IP address in ARP/RARP packet
        comment: string, text with max. 256 characters
        gratuitous: string, boolean indicating whether to check for gratuitous ARP packet
        """

        __slots__ = ('srcmacaddr', 'srcmacmask', 'dstmacaddr', 'dstmacmask',
                     'hwtype', 'protocoltype', 'opcode', 'arpsrcmacaddr',
                     'arpdstmacaddr', 'arpsrcipaddr', 'arpdstipaddr',
                     'comment', 'gratuitous')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLAttribute('srcmacaddr', self, parent_xpath='/',
                                   tag_name='arp', attribute='srcmacaddr')
            accessors.XMLAttribute('srcmacmask', self, parent_xpath='/',
                                   tag_name='arp', attribute='srcmacmask')
            accessors.XMLAttribute('dstmacaddr', self, parent_xpath='/',
                                   tag_name='arp', attribute='dstmacaddr')
            accessors.XMLAttribute('dstmacmask', self, parent_xpath='/',
                                   tag_name='arp', attribute='dstmacmask')
            accessors.XMLAttribute('hwtype', self, parent_xpath='/',
                                   tag_name='arp', attribute='hwtype')
            accessors.XMLAttribute('protocoltype', self, parent_xpath='/',
                                   tag_name='arp', attribute='protocoltype')
            accessors.XMLAttribute('opcode', self, parent_xpath='/',
                                   tag_name='arp', attribute='opcode')
            accessors.XMLAttribute('arpsrcmacaddr', self, parent_xpath='/',
                                   tag_name='arp', attribute='arpsrcmacaddr')
            accessors.XMLAttribute('arpdstmacaddr', self, parent_xpath='/',
                                   tag_name='arp', attribute='arpdstmacaddr')
            accessors.XMLAttribute('arpsrcipaddr', self, parent_xpath='/',
                                   tag_name='arp', attribute='arpsrcipaddr')
            accessors.XMLAttribute('arpdstipaddr', self, parent_xpath='/',
                                   tag_name='arp', attribute='arpdstipaddr')
            accessors.XMLAttribute('comment', self, parent_xpath='/',
                                   tag_name='arp', attribute='comment')
            accessors.XMLAttribute('gratuitous', self, parent_xpath='/',
                                   tag_name='arp', attribute='gratuitous')

            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<arp/>'



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2013-2014
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

import sys

# pylint: disable=E0611
from setuptools import setup, find_packages

VERSION = open('VERSION', 'r').read().strip()


def pre_post_plugin_type():
    try:
        from avocado.core.plugin_interfaces import JobPreTests as Pre
        return 'avocado.plugins.result_events'
    except ImportError:
        return 'avocado.plugins.job.prepost'


if __name__ == "__main__":
    requirements = ["netifaces", "simplejson", "six"]
    if sys.version_info[:2] >= (3, 0):
        requirements.append("avocado-framework>=68.0")
        requirements.append("netaddr")
        requirements.append("aexpect")
    else:
        # Latest py2 supported stevedore is 1.10.0, need to limit it here
        # as older avocado versions were not limiting it.
        # Note: Avocado 70+ doesn't require stevedore and older Avocado
        # can use whatever version of stevedore on py3
        requirements.append("aexpect<=1.6.0")
        requirements.append("urllib3<=1.24.3")
        requirements.append("stevedore>=1.8.0,<=1.10.0")
        requirements.append("avocado-framework>=68.0,<70.0")
        requirements.append("netaddr<=0.7.19")

    setup(name='avocado-framework-plugin-vt',
          version=VERSION,
          description='Avocado Plugin for Virtualization Testing',
          author='Avocado Developers',
          author_email='avocado-devel@redhat.com',
          url='http://github.com/avocado-framework/avocado-vt',
          packages=find_packages(exclude=('selftests*',)),
          include_package_data=True,
          entry_points={
              'avocado.plugins.settings': [
                  'vt-settings = avocado_vt.plugins.vt_settings:VTSettings',
                  ],
              'avocado.plugins.cli': [
                  'vt-list = avocado_vt.plugins.vt_list:VTLister',
                  'vt = avocado_vt.plugins.vt:VTRun',
                  ],
              'avocado.plugins.cli.cmd': [
                  'vt-bootstrap = avocado_vt.plugins.vt_bootstrap:VTBootstrap',
                  ],
              pre_post_plugin_type(): [
                  'vt-joblock = avocado_vt.plugins.vt_joblock:VTJobLock',
                  ],
              'avocado.plugins.init': [
                  'vt-init = avocado_vt.plugins.vt_init:VtInit',
                  ],
              },
          install_requires=requirements,
          )



import logging
import re

from avocado.utils import path
from avocado.utils import process
from avocado.core import exceptions

from virttest import error_context
from virttest import utils_test

# command `grep --color` may have alias name `grep` in some systems,
# so get explicit command 'grep' with path
grep_binary = path.find_command("grep")


@error_context.context_aware
def get_host_timezone():
    """
    Get host's timezone
    """
    timezone_cmd = 'timedatectl | %s "Time zone"' % grep_binary
    timezone_pattern = '^(?:\s+Time zone:\s)(\w+\/\S+|UTC)(?:\s\(\S+,\s)([+|-]\d{4})\)$'
    error_context.context("Get host's timezone", logging.info)
    host_timezone = process.run(timezone_cmd, timeout=240, shell=True).stdout_text
    try:
        host_timezone_set = re.match(timezone_pattern, host_timezone).groups()
        return {"timezone_city": host_timezone_set[0],
                "timezone_code": host_timezone_set[1]}
    except (AttributeError, IndexError):
        raise exceptions.TestError("Fail to get host's timezone.")


@error_context.context_aware
def verify_timezone_linux(session):
    """
    Verify linux guest's timezone

    :param session: VM session
    """
    error_context.context("Verify guest's timezone", logging.info)
    timezone_cmd = 'timedatectl | %s "Time zone"' % grep_binary
    timezone_pattern = '(?:\s+Time zone:\s)(\w+\/\S+|UTC)(?:\s\(\S+,\s)([+|-]\d{4})\)'
    guest_timezone = session.cmd_output_safe(timezone_cmd, timeout=240)
    try:
        guest_timezone_set = re.search(timezone_pattern, guest_timezone).groups()
        return guest_timezone_set[0] == get_host_timezone()['timezone_city']
    except (AttributeError, IndexError):
        raise exceptions.TestError("Fail to get guest's timezone.")


@error_context.context_aware
def sync_timezone_linux(vm):
    """
    Sync linux guest's timezone

    :param vm: Virtual machine object
    """
    session = vm.wait_for_login(serial=True)
    error_context.context("Sync guest's timezone", logging.info)
    set_timezone_cmd = "timedatectl set-timezone %s"
    if not verify_timezone_linux(session):
        host_timezone_city = get_host_timezone()['timezone_city']
        session.cmd(set_timezone_cmd % host_timezone_city)
        if not verify_timezone_linux(session):
            session.close()
            raise exceptions.TestError("Fail to sync guest's timezone.")
    session.close()


@error_context.context_aware
def verify_timezone_win(session):
    """
    Verify windows guest's timezone

    :params session: VM session
    :return tuple(verify_status, get_timezone_name)
    """
    def get_timezone_list():
        timezone_list_cmd = "tzutil /l"
        timezone_list = session.cmd_output_safe(timezone_list_cmd)

        match_pattern = "(?:\(UTC([+|-]\d{2}:\d{2})?)(?:.*\n)(\w+.*(?:\s\w+)*)"
        timezone_sets = []
        for para in re.split("(?:\r?\n){2,}", timezone_list.strip()):
            result = re.match(match_pattern, para, re.M)
            if not result:
                continue
            code, name = result.groups()
            # When UTC standard time, add timezone code '+0000'
            if not code:
                code = "+0000"
            else:
                code = re.sub(r'(\d{2}):(\d{2})', r'\1\2', code)
            timezone_sets.append((code, name))
        return timezone_sets

    def get_timezone_code(timezone_name):
        for value in get_timezone_list():
            if value[1] == timezone_name:
                return value[0]
        return None

    def get_timezone_name(timezone_code):
        for value in get_timezone_list():
            if value[0] == timezone_code:
                return value[1]
        return None

    error_context.context("Verify guest's timezone", logging.info)
    timezone_cmd = 'tzutil /g'
    host_timezone_code = get_host_timezone()['timezone_code']
    # Workaround to handle two line prompts in serial session
    timezone_name = session.cmd_output_safe(timezone_cmd).split('\n')[0]
    if get_timezone_code(timezone_name) != host_timezone_code:
        return False, get_timezone_name(host_timezone_code)
    return True, ""


@error_context.context_aware
def sync_timezone_win(vm):
    """
    Verify and sync windows guest's timezone

    :param vm: Virtual machine object
    """
    session = vm.wait_for_login(serial=True)
    set_timezone_cmd = 'tzutil /s "%s"'
    (ver_result, output) = verify_timezone_win(session)

    if ver_result is not True:
        error_context.context("Sync guest's timezone.", logging.info)
        session.cmd(set_timezone_cmd % output)
        vm_params = vm.params
        error_context.context("Shutdown guest...", logging.info)
        vm.destroy()
        error_context.context("Boot guest...", logging.info)
        vm.create(params=vm_params)
        vm.verify_alive()
        session = vm.wait_for_login(serial=True)
        (ver_result, output) = verify_timezone_win(session)
        if ver_result is not True:
            session.close()
            raise exceptions.TestError("Fail to sync guest's timezone.")
    session.close()


def execute(cmd, timeout=360, session=None):
    """
    Execute command in guest or host, if session is not None return
    command output in guest else return command ouput in host

    :param cmd: Shell commands
    :param timeout: Timeout to execute command
    :param session: ShellSession or None

    :return: Command output string
    """
    if session:
        ret = session.cmd_output_safe(cmd, timeout=timeout)
    else:
        ret = process.getoutput(cmd)
    target = 'guest' if session else 'host'
    logging.debug("(%s) Execute command('%s')" % (target, cmd))
    return ret


@error_context.context_aware
def verify_clocksource(expected, session=None):
    """
    Verify if host/guest use the expected clocksource
    :param expected: Expected clocksource
    :param session: VM session
    """
    error_context.context("Check the current clocksource", logging.info)
    cmd = "cat /sys/devices/system/clocksource/"
    cmd += "clocksource0/current_clocksource"
    return expected in execute(cmd, session=session)


@error_context.context_aware
def sync_time_with_ntp(session=None):
    """
    Sync guest or host time with ntp server
    :param session: VM session or None
    """
    error_context.context("Sync time from ntp server", logging.info)
    cmd = "ntpdate clock.redhat.com; hwclock -w"
    return execute(cmd, session)


@error_context.context_aware
def update_clksrc(vm, clksrc=None):
    """
    Update linux guest's clocksource and re-boot guest

    :params vm: Virtual machine for vm
    :params clksrc: Expected clocksource
    """
    params = vm.get_params()
    if 'fedora' in params["os_variant"] and clksrc and clksrc != 'kvm-clock':
        cpu_model_flags = params.get["cpu_model_flags"]
        params["cpu_model_flags"] = cpu_model_flags + ",-kvmclock"

    error_context.context("Update guest kernel cli to '%s'" %
                          (clksrc or "kvm-clock"),
                          logging.info)
    if clksrc:
        boot_option_added = "clocksource=%s" % clksrc
        utils_test.update_boot_option(vm, args_added=boot_option_added)


def is_ntp_enabled(session):
    """
    Get current NTP state for guest/host
    """
    cmd = 'timedatectl | %s "NTP enabled"' % grep_binary
    return 'yes' in execute(cmd, session=session).split(":")[1].strip()


def ntp_switch(session, off=True):
    """
    Turn off/on ntp for guest/host
    """
    cmd = "timedatectl set-ntp"
    if off:
        cmd += " 0"
    else:
        cmd += " 1"
    output = execute(cmd, session=session)
    ntp_enabled = is_ntp_enabled(session)
    if off and ntp_enabled:
        raise exceptions.TestError("Fail to switchoff ntp: %s", output)
    if not off and not ntp_enabled:
        raise exceptions.TestError("Fail to switchon ntp: %s", output)



"""
Classes and functions to handle storage devices.

This exports:
  - two functions for get image/blkdebug filename
  - class for image operates and basic parameters
"""
from __future__ import division
import errno
import logging
import os
import shutil
import re
import functools
import collections
import json

from avocado.core import exceptions
from avocado.utils import process

from virttest import storage_ssh
from virttest import nbd
from virttest import curl
from virttest import iscsi
from virttest import utils_misc
from virttest import utils_numeric
from virttest import utils_params
from virttest import virt_vm
from virttest import gluster
from virttest import lvm
from virttest import ceph
from virttest import nvme
from virttest import data_dir


def preprocess_images(bindir, params, env):
    # Clone master image form vms.
    for vm_name in params.get("vms").split():
        vm = env.get_vm(vm_name)
        if vm:
            vm.destroy(free_mac_addresses=False)
        vm_params = params.object_params(vm_name)
        for image in vm_params.get("master_images_clone").split():
            image_obj = QemuImg(vm_params, bindir, image)
            image_obj.clone_image(vm_params, vm_name, image, bindir)


def preprocess_image_backend(bindir, params, env):
    enable_gluster = params.get("enable_gluster") == "yes"
    gluster_image = params.get("gluster_brick")
    if enable_gluster and gluster_image:
        return gluster.create_gluster_vol(params)

    return True


def postprocess_images(bindir, params):
    for vm in params.get("vms").split():
        vm_params = params.object_params(vm)
        for image in vm_params.get("master_images_clone").split():
            image_obj = QemuImg(vm_params, bindir, image)
            image_obj.rm_cloned_image(vm_params, vm, image, bindir)


def file_exists(params, filename_path):
    """
    Check if image_filename exists.

    :param params: Dictionary containing the test parameters.
    :param filename_path: path to file
    :type filename_path: str
    :param root_dir: Base directory for relative filenames.
    :type root_dir: str

    :return: True if image file exists else False
    """
    if params.get("enable_gluster") == "yes":
        return gluster.file_exists(params, filename_path)

    if params.get("enable_ceph") == "yes":
        image_name = params.get("image_name")
        image_format = params.get("image_format", "qcow2")
        ceph_monitor = params.get("ceph_monitor")
        rbd_pool_name = params["rbd_pool_name"]
        rbd_namespace_name = params.get("rbd_namespace_name")
        rbd_image_name = "%s.%s" % (image_name.split("/")[-1], image_format)
        ceph_conf = params.get("ceph_conf")
        keyring_conf = params.get("image_ceph_keyring_conf")
        return ceph.rbd_image_exist(ceph_monitor, rbd_pool_name, rbd_image_name,
                                    ceph_conf, keyring_conf, rbd_namespace_name)

    if params.get('enable_nvme') == 'yes':
        return nvme.file_exists(params, filename_path)

    if params.get('enable_ssh') == 'yes':
        return storage_ssh.file_exists(params, filename_path)

    if params.get('enable_curl') == 'yes':
        return curl.file_exists(params, filename_path)

    return os.path.exists(filename_path)


def file_remove(params, filename_path):
    """
    Remove the image
    :param params: Dictionary containing the test parameters.
    :param filename_path: path to file
    """
    if params.get("enable_ceph") == "yes":
        image_name = params.get("image_name")
        image_format = params.get("image_format", "qcow2")
        ceph_monitor = params.get("ceph_monitor")
        rbd_pool_name = params["rbd_pool_name"]
        rbd_namespace_name = params.get("rbd_namespace_name")
        rbd_image_name = "%s.%s" % (image_name.split("/")[-1], image_format)
        ceph_conf = params.get("ceph_conf")
        keyring_conf = params.get("image_ceph_keyring_conf")
        return ceph.rbd_image_rm(ceph_monitor, rbd_pool_name, rbd_image_name,
                                 ceph_conf, keyring_conf, rbd_namespace_name)

    if params.get("gluster_brick"):
        # TODO: Add implementation for gluster_brick
        return

    if params.get('storage_type') in ('iscsi', 'lvm', 'iscsi-direct'):
        # TODO: Add implementation for iscsi/lvm
        return

    # skip removing raw device
    if params.get('image_raw_device') == 'yes':
        return

    if os.path.exists(filename_path):
        os.unlink(filename_path)
        return


def get_image_blkdebug_filename(params, root_dir):
    """
    Generate an blkdebug file path from params and root_dir.

    blkdebug files allow error injection in the block subsystem.

    :param params: Dictionary containing the test parameters.
    :param root_dir: Base directory for relative filenames.

    :note: params should contain:
           blkdebug -- the name of the debug file.
    """
    blkdebug_name = params.get("drive_blkdebug", None)
    if blkdebug_name is not None:
        blkdebug_filename = utils_misc.get_path(root_dir, blkdebug_name)
    else:
        blkdebug_filename = None
    return blkdebug_filename


def get_image_filename(params, root_dir, basename=False):
    """
    Generate an image path from params and root_dir.

    :param params: Dictionary containing the test parameters.
    :param root_dir: Base directory for relative filenames.
    :param basename: True to use only basename of image_name

    :note: params should contain:
           image_name -- the name of the image file, without extension
           image_format -- the format of the image (qcow2, raw etc)
    :raise VMDeviceError: When no matching disk found (in indirect method).
    """
    enable_ssh = params.get("enable_ssh") == "yes"
    enable_curl = params.get("enable_curl") == "yes"
    enable_nbd = params.get("enable_nbd") == "yes"
    enable_gluster = params.get("enable_gluster") == "yes"
    enable_ceph = params.get("enable_ceph") == "yes"
    enable_iscsi = params.get("enable_iscsi") == "yes"
    enable_nvme = params.get("enable_nvme") == "yes"
    image_name = params.get("image_name")
    storage_type = params.get("storage_type")
    if image_name:
        image_format = params.get("image_format", "qcow2")
        if enable_curl:
            # required libcurl params
            curl_protocol = params['curl_protocol']
            curl_server = params['curl_server']
            curl_path = params['curl_path']

            # optional libcurl params
            curl_user = params.get('curl_username')
            curl_passwd = params.get('curl_password')

            return curl.get_image_filename(curl_protocol, curl_server,
                                           curl_path, curl_user, curl_passwd)
        if enable_nbd:
            return nbd.get_image_filename(params)
        if enable_iscsi:
            if storage_type == 'iscsi-direct':
                portal = params.get('portal_ip')
                target = params.get('target')
                lun = params.get('lun', 0)
                user = params.get('chap_user')
                password = params.get('chap_passwd')
                return iscsi.get_image_filename(portal, target, lun,
                                                user, password)
        if enable_gluster:
            return gluster.get_image_filename(params, image_name, image_format)
        if enable_ceph:
            rbd_pool_name = params["rbd_pool_name"]
            rbd_namespace_name = params.get("rbd_namespace_name")
            rbd_image_name = "%s.%s" % (image_name.split("/")[-1],
                                        image_format)
            ceph_conf = params.get('ceph_conf')
            ceph_monitor = params.get('ceph_monitor')
            return ceph.get_image_filename(ceph_monitor, rbd_pool_name,
                                           rbd_image_name, ceph_conf,
                                           rbd_namespace_name)
        if enable_nvme:
            address = params['nvme_pci_address']
            namespace = params.get('nvme_namespace', 1)
            return nvme.get_image_filename(address, namespace)
        if enable_ssh:
            # required libssh options
            server = params['image_ssh_host']
            ssh_image_path = params['image_ssh_path']

            # optional libssh options
            user = params.get('image_ssh_user')
            port = params.get('image_ssh_port')
            host_key_check = params.get('image_ssh_host_key_check')

            return storage_ssh.get_image_filename(server, ssh_image_path,
                                                  user, port, host_key_check)
        return get_image_filename_filesytem(params, root_dir, basename=basename)
    else:
        logging.warn("image_name parameter not set.")


def get_image_filename_filesytem(params, root_dir, basename=False):
    """
    Generate an image path from params and root_dir.

    :param params: Dictionary containing the test parameters.
    :param root_dir: Base directory for relative filenames.
    :param basename: True to use only basename of image_name

    :note: params should contain:
           image_name -- the name of the image file, without extension
           image_format -- the format of the image (qcow2, raw etc)
    :raise VMDeviceError: When no matching disk found (in indirect method).
    """
    def sort_cmp(first, second):
        """
        This function used for sort to suit for this test, first sort by len
        then by value.
        """
        def cmp(x, y):
            return (x > y) - (x < y)

        first_contains_digit = re.findall(r'[vhs]d[a-z]*[\d]+', first)
        second_contains_digit = re.findall(r'[vhs]d[a-z]*[\d]+', second)

        if not first_contains_digit and not second_contains_digit:
            if len(first) > len(second):
                return 1
            elif len(first) < len(second):
                return -1
        if len(first) == len(second):
            if first_contains_digit and second_contains_digit:
                return cmp(first, second)
            elif first_contains_digit:
                return -1
            elif second_contains_digit:
                return 1
        return cmp(first, second)

    image_name = params.get("image_name", "image")
    if basename:
        image_name = os.path.basename(image_name)
    indirect_image_select = params.get("indirect_image_select")
    if indirect_image_select:
        re_name = image_name
        indirect_image_select = int(indirect_image_select)
        matching_images = process.run("ls -1d %s" % re_name,
                                      shell=True).stdout_text
        matching_images = sorted(matching_images.split('\n'),
                                 key=functools.cmp_to_key(sort_cmp))
        if matching_images[-1] == '':
            matching_images = matching_images[:-1]
        try:
            image_name = matching_images[indirect_image_select]
        except IndexError:
            raise virt_vm.VMDeviceError("No matching disk found for "
                                        "name = '%s', matching = '%s' and "
                                        "selector = '%s'" %
                                        (re_name, matching_images,
                                         indirect_image_select))
        for protected in params.get('indirect_image_blacklist', '').split(' '):
            match_image = re.match(protected, image_name)
            if match_image and match_image.group(0) == image_name:
                # We just need raise an error if it is totally match, such as
                # sda sda1 and so on, but sdaa should not raise an error.
                raise virt_vm.VMDeviceError("Matching disk is in blacklist. "
                                            "name = '%s', matching = '%s' and "
                                            "selector = '%s'" %
                                            (re_name, matching_images,
                                             indirect_image_select))

    image_format = params.get("image_format", "qcow2")
    if params.get("image_raw_device") == "yes":
        return image_name
    if image_format:
        image_filename = "%s.%s" % (image_name, image_format)
    else:
        image_filename = image_name

    image_filename = utils_misc.get_path(root_dir, image_filename)
    return image_filename


def get_iso_filename(cdrom_params, root_dir, basename=False):
    """
    Generate an iso image path from params and root_dir.

    :param cdrom_params: Dictionary containing the test parameters.
    :param root_dir: Base directory for relative iso image.
    :param basename: True to use only basename of iso image.
    :return: iso filename
    """
    enable_nvme = cdrom_params.get("enable_nvme") == "yes"
    enable_nbd = cdrom_params.get("enable_nbd") == "yes"
    enable_gluster = cdrom_params.get("enable_gluster") == "yes"
    enable_ceph = cdrom_params.get("enable_ceph") == "yes"
    enable_iscsi = cdrom_params.get("enable_iscsi") == "yes"
    enable_curl = cdrom_params.get("enable_curl") == "yes"
    enable_ssh = cdrom_params.get("enable_ssh") == "yes"

    if enable_nvme:
        return None
    elif any((enable_nbd, enable_gluster, enable_ceph,
              enable_iscsi, enable_curl, enable_ssh)):
        return get_image_filename(cdrom_params, None, basename)
    else:
        iso = cdrom_params.get("cdrom")
        if iso:
            iso = os.path.basename(iso) if basename else utils_misc.get_path(
                                                               root_dir, iso)
        return iso


secret_dir = os.path.join(data_dir.get_data_dir(), "images/secrets")


def _make_secret_dir():
    """Create image secret directory."""
    try:
        os.makedirs(secret_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


class ImageSecret(object):
    """Image secret object."""

    def __init__(self, image, data):
        if not data:
            raise ValueError("Empty image secret for %s" % image)
        self.image_id = image
        self.data = data
        self.filename = os.path.join(secret_dir, "%s.secret" % image)
        self.aid = "%s_%s" % (self.image_id, "encrypt0")

    def __str__(self):
        return self.aid

    @classmethod
    def image_secret_define_by_params(cls, image, params):
        """Get image secret from vt params."""
        image_secret = params.get("image_secret")
        image_format = params.get("image_format")
        image_encryption = params.get("image_encryption")
        if ((image_format == "qcow2" and image_encryption == "luks") or
                image_format == "luks"):
            return cls(image, image_secret)

    def save_to_file(self):
        """Save secret data to file."""
        _make_secret_dir()
        with open(self.filename, "w") as fd:
            fd.write(self.data)


class Cookie(object):
    """
    Cookie data stored in secret object
    """

    def __init__(self, image, cookie_data, cookie_data_format):
        """
        :param image: image tag name
        :param cookie_data: cookie data string
        :param cookie_data_format: raw or base64
        """
        self.image = image
        self.data = cookie_data
        self.data_format = cookie_data_format
        self.aid = '%s_cookie_secret' % self.image
        self.filename = os.path.join(secret_dir, "%s.secret" % self.aid)
        self.save_to_file()

    def save_to_file(self):
        """Save secret data to file."""
        _make_secret_dir()
        with open(self.filename, "w") as fd:
            fd.write(self.data)


class StorageAuth(object):
    """
    Image storage authentication class.
    iscsi auth: initiator + password
    ceph auth: ceph key
    nbd auth: tls creds
    libcurl auth: password sslverify timeout readahead cookie
    """

    def __init__(self, image, data, data_format, storage_type, **info):
        """
        :param image: image tag name
        :param data: sensitive data like password
        :param data_format: raw or base64
        :param storage_type: ceph, glusterfs-direct, iscsi-direct, nbd, curl
        :param info: other access information, such as:
                     iscsi-direct: initiator
                     gluster-direct: debug, logfile, peers
                     nbd: tls creds path for client access
                     libcurl: password, sslverify, timeout, readahead, cookie
        """
        self.image = image
        self.aid = '%s_access' % self.image
        self.storage_type = storage_type
        self.filename = os.path.join(secret_dir, "%s.secret" % self.aid)
        self.data_format = data_format

        if self.storage_type == 'iscsi-direct':
            self._chap_passwd = data
            self.iscsi_initiator = info.get('initiator')
        elif self.storage_type == 'ceph':
            self._ceph_key = data
        elif self.storage_type == 'glusterfs-direct':
            self.peers = info['peers']

            # TODO: logfile and debug will be moved to a class,
            # they'll be defined as common options for all backends
            self.debug = info['debug']
            self.logfile = info['logfile']
        elif self.storage_type == 'nbd':
            # TODO: now we only support tls-creds-x509, we can add a
            # param 'nbd_tls_creds_object' to differentiate objects,
            # e.g. tls-creds-psk
            self.tls_creds = info['tls_creds']
            self.reconnect_delay = info['reconnect_delay']
        elif self.storage_type == 'curl':
            self._password = data
            self.cookie = info['cookie']
            self.sslverify = info['sslverify']
            self.readahead = info['readahead']
            self.timeout = info['timeout']

        if self.data is not None:
            self.filename = os.path.join(secret_dir, "%s.secret" % self.aid)
            self.save_to_file()

    @property
    def data(self):
        if self.storage_type == 'iscsi-direct':
            return self._chap_passwd
        elif self.storage_type == 'ceph':
            return self._ceph_key
        elif self.storage_type == 'curl':
            return self._password
        else:
            return None

    def save_to_file(self):
        """Save secret data to file."""
        _make_secret_dir()
        with open(self.filename, "w") as fd:
            fd.write(self.data)

    @classmethod
    def auth_info_define_by_params(cls, image, params):
        """
        :param image: image tag name
        :param params: image specified parmas, i.e. params.object_params(image)
        """
        auth = None
        storage_type = params.get("storage_type")
        enable_ceph = params.get("enable_ceph") == "yes"
        enable_iscsi = params.get("enable_iscsi") == "yes"
        enable_gluster = params.get("enable_gluster") == "yes"
        enable_nbd = params.get("enable_nbd") == "yes"
        enable_curl = params.get("enable_curl") == "yes"

        if enable_iscsi:
            if storage_type == 'iscsi-direct':
                initiator = params.get('initiator')
                data = params.get('chap_passwd')
                data_format = params.get('data_format', 'raw')
                auth = cls(image, data, data_format, storage_type,
                           initiator=initiator) if data or initiator else None
        elif enable_ceph:
            data = params.get('ceph_key')
            data_format = params.get('data_format', 'base64')
            auth = cls(image, data, data_format,
                       storage_type) if data else None
        elif enable_gluster:
            if storage_type == 'glusterfs-direct':
                peers = json.loads(params.get('gluster_peers', '[]'))
                debug = params.get('gluster_debug')
                logfile = params.get('gluster_logfile')
                auth = cls(image, None, None, storage_type,
                           debug=debug, logfile=logfile,
                           peers=peers) if debug or logfile or peers else None
        elif enable_nbd:
            # tls-creds is supported for ip only
            tls_creds = params.get(
                'nbd_client_tls_creds') if params.get('nbd_server') else None
            reconnect_delay = params.get('nbd_reconnect_delay')
            auth = cls(
                image, None, None, storage_type,
                tls_creds=tls_creds, reconnect_delay=reconnect_delay
            ) if tls_creds or reconnect_delay else None
        elif enable_curl:
            # cookie data in a secure way, only for http/https
            cookie = Cookie(
                image, params['curl_cookie_secret'],
                params.get('curl_cookie_secret_format', 'raw'),
            ) if params.get('curl_cookie_secret') else None

            # sslverify, only for https/ftps
            sslverify = params.get('curl_sslverify')

            # size of the read-ahead cache
            readahead = int(float(
                utils_numeric.normalize_data_size(params['curl_readahead'],
                                                  order_magnitude="B")
            )) if params.get('curl_readahead') else None

            # timeout for connections in seconds
            timeout = params.get('curl_timeout')

            # password
            data = params.get('curl_password')
            data_format = params.get('curl_password_format', 'raw')

            if any((data, cookie, sslverify, readahead, timeout)):
                auth = cls(image, data, data_format, storage_type,
                           cookie=cookie, sslverify=sslverify,
                           readahead=readahead, timeout=timeout)

        return auth


class ImageAccessInfo(object):
    """
    Access info to the logical image, which can include the network
    storage image only, or the image and its backing images.
    """
    def __init__(self, image, image_auth, image_backing_auth):
        """
        :param image: image tag name
        :param image_auth: StorageAuth object to access image itself
        :param image_backing_auth: a dict({image: StorageAuth object}),
                                   used for accessing the backing images
        """
        self.image = image
        self.image_auth = image_auth
        self.image_backing_auth = image_backing_auth

    @classmethod
    def access_info_define_by_params(cls, image, params):
        """
        :param image: image tag name
        :param params: a dict containing the test parameters

        :return: an ImageAccessInfo object or None
        """
        access_info = None
        info = retrieve_access_info(image, params)

        if info:
            access = info.pop(image, None)
            access_info = cls(image, access, info)

        return access_info


def retrieve_access_info(image, params):
    """
    Create the image and its backing images' access info objects,
    keep the same order as the images in the image_chain.

    :param image: image tag name
    :param params: a dict containing the test parameters

    :return: A dict({image: StorageAuth object or None})
    """
    access_info = collections.OrderedDict()
    image_chain = params.objects("image_chain")
    images = image_chain if image_chain else [image]

    for img in images:
        auth = StorageAuth.auth_info_define_by_params(
                       img, params.object_params(img))
        if auth is not None:
            access_info[img] = auth

        if img == image:
            # ignore image's snapshot images
            break

    return access_info


def retrieve_secrets(image, params):
    """Get all image secrets in image_chain, up to image."""
    secrets = []
    # use image instead if image_chain is empty
    # or no backing image available
    image_chain = params.get("image_chain", "")
    if image not in image_chain:
        image_chain = image
    for img in image_chain.split():
        img_params = params.object_params(img)
        secret = ImageSecret.image_secret_define_by_params(img, img_params)
        if secret:
            secrets.append(secret)
        # NOTE: break here to exclude secrets of snapshots.
        if img == image:
            break
    return secrets


class ImageEncryption(object):
    """Image encryption configuration."""

    __slots__ = ("format", "key_secret", "base_key_secrets",
                 "cipher_alg", "cipher_mode", "ivgen_alg",
                 "ivgen_hash_alg", "hash_alg", "iter_time")

    def __init__(self, encryption_format, key_secret, base_key_secrets,
                 cipher_alg, cipher_mode, ivgen_alg, ivgen_hash_alg, hash_alg,
                 iter_time):
        """
        Initialize image encryption configuration.

        :param encrypt_format: encryption format
        :param key_secret: ImageSecret object for this image
        :param base_key_secrets: ImageSecret objects from its backing images
        :param cipher_alg: name of encryption cipher algorithm
        :param cipher_mode: name of encryption cipher mode
        :param ivgen_alg: name of iv generator algorithm
        :param ivgen_hash_alg: name of iv generator hash algorithm
        :param hash_alg: name of encryption hash algorithm
        :param iter_time: time to spend in PBKDF in milliseconds
        """
        self.format = encryption_format
        self.key_secret = key_secret
        self.base_key_secrets = base_key_secrets
        self.cipher_alg = cipher_alg
        self.cipher_mode = cipher_mode
        self.ivgen_alg = ivgen_alg
        self.ivgen_hash_alg = ivgen_hash_alg
        self.hash_alg = hash_alg
        self.iter_time = iter_time

    def __iter__(self):
        return iter(self.__slots__)

    @classmethod
    def encryption_define_by_params(cls, image, params):
        """Get image encryption from vt params."""
        encryption_format = params.get("image_encryption")
        key_secrets = retrieve_secrets(image, params)
        key_secret = None
        if key_secrets and key_secrets[-1].image_id == image:
            key_secret = key_secrets.pop()
        cipher_alg = params.get("image_cipher_alg")
        cipher_mode = params.get("image_cipher_mode")
        ivgen_alg = params.get("image_ivgen_alg")
        ivgen_hash_alg = params.get("image_ivgen_hash_alg")
        hash_alg = params.get("image_hash_alg")
        iter_time = params.get("image_iter_time")
        return cls(encryption_format, key_secret, key_secrets, cipher_alg,
                   cipher_mode, ivgen_alg, ivgen_hash_alg, hash_alg, iter_time)

    @property
    def image_key_secrets(self):
        """All image secrets required to use this image."""
        if self.key_secret:
            return self.base_key_secrets + [self.key_secret]
        return self.base_key_secrets


def copy_nfs_image(params, root_dir, basename=False):
    """
    copy image from image_path to nfs mount dir if image is not available
    or corrupted.

    :param params: Test dict params
    :param root_dir: Base directory for relative filenames.
    :param basename: True to use only basename of image name
    :raise: TestSetupFail if image is unavailable/corrupted
    """
    if params.get("setup_local_nfs", "no") == "yes":
        # check for image availability in NFS shared path
        base_dir = params["nfs_mount_dir"]
        dst = get_image_filename(params, base_dir, basename=basename)
        if(not os.path.isfile(dst) or
           utils_misc.get_image_info(dst)['lcounts'].lower() == "true"):
            source = get_image_filename(params, root_dir)
            logging.debug("Checking for image available in image data "
                          "path - %s", source)
            # check for image availability in images data directory
            if(os.path.isfile(source) and not
               utils_misc.get_image_info(source)['lcounts'].lower() == "true"):
                logging.debug("Copying guest image from %s to %s", source,
                              dst)
                shutil.copy(source, dst)
            else:
                raise exceptions.TestSetupFail("Guest image is unavailable"
                                               "/corrupted in %s and %s" %
                                               (source, dst))


class OptionMissing(Exception):

    """
    Option not found in the odbject
    """

    def __init__(self, option):
        self.option = option

    def __str__(self):
        return "%s is missing. Please check your parameters" % self.option


class QemuImg(object):

    """
    A basic class for handling operations of disk/block images.
    """

    def __init__(self, params, root_dir, tag):
        """
        Init the default value for image object.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param tag: Image tag defined in parameter images.
        """
        self.params = params
        self.image_filename = get_image_filename(params, root_dir)
        self.image_format = params.get("image_format", "qcow2")
        self.size = params.get("image_size", "10G")
        self.storage_type = params.get("storage_type", "local fs")
        self.check_output = params.get("check_output") == "yes"
        self.image_blkdebug_filename = get_image_blkdebug_filename(params,
                                                                   root_dir)
        self.remote_keywords = params.get(
            "remote_image",
            "gluster iscsi rbd nbd nvme http https ftp ftps"
        ).split()
        self.encryption_config = ImageEncryption.encryption_define_by_params(
            tag, params)
        image_chain = params.get("image_chain")
        self.tag = tag
        self.root_dir = root_dir
        self.base_tag = None
        self.snapshot_tag = None
        if image_chain:
            image_chain = re.split(r"\s+", image_chain)
            if tag in image_chain:
                index = image_chain.index(tag)
                if index < len(image_chain) - 1:
                    self.snapshot_tag = image_chain[index + 1]
                if index > 0:
                    self.base_tag = image_chain[index - 1]
        if self.base_tag:
            base_params = params.object_params(self.base_tag)
            self.base_image_filename = get_image_filename(base_params,
                                                          root_dir)
            self.base_format = base_params.get("image_format")
        if self.snapshot_tag:
            ss_params = params.object_params(self.snapshot_tag)
            self.snapshot_image_filename = get_image_filename(ss_params,
                                                              root_dir)
            self.snapshot_format = ss_params.get("image_format")

        self.image_access = ImageAccessInfo.access_info_define_by_params(
                                                   self.tag, self.params)

        self.data_file = self.external_data_file_defined_by_params(
            params, root_dir, tag)

    @classmethod
    def external_data_file_defined_by_params(cls, params, root_dir, tag):
        """Link image to an external data file."""
        enable_data_file = params.get("enable_data_file", "no") == "yes"
        image_format = params.get("image_format", "qcow2")
        if not enable_data_file:
            return
        if image_format != "qcow2":
            raise ValueError("The %s format does not support external "
                             "data file" % image_format)
        image_size = params["image_size"]
        base_name = os.path.basename(params["image_name"])
        data_file_path = params.get("image_data_file_path",
                                    os.path.join(root_dir, "images",
                                                 "%s.data_file"
                                                 % base_name))
        data_file_params = utils_params.Params(
            {"image_name": data_file_path,
             "image_format": "raw",
             "image_size": image_size,
             "image_raw_device": "yes"})
        return cls(data_file_params, root_dir, "%s_data_file" % tag)

    def check_option(self, option):
        """
        Check if object has the option required.

        :param option: option should be checked
        """
        if option not in self.__dict__:
            raise OptionMissing(option)

    def is_remote_image(self):
        """
        Check if image is from a remote server or not
        """

        for keyword in self.remote_keywords:
            if self.image_filename.startswith(keyword):
                return True

        return False

    def backup_image(self, params, root_dir, action, good=True,
                     skip_existing=False):
        """
        Backup or restore a disk image, depending on the action chosen.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param action: Whether we want to backup or restore the image.
        :param good: If we are backing up a good image(we want to restore it)
            or a bad image (we are saving a bad image for posterior analysis).

        :note: params should contain:
               image_name -- the name of the image file, without extension
               image_format -- the format of the image (qcow2, raw etc)
        """
        def get_backup_set(filename, backup_dir, action, good):
            """
            Get all sources and destinations required for each backup.
            """
            if not os.path.isdir(backup_dir):
                os.makedirs(backup_dir)
            basename = os.path.basename(filename)
            bkp_set = []
            if action not in ('backup', 'restore'):
                logging.error("No backup sets for action: %s, state: %s",
                              action, good)
                return bkp_set
            if good:
                src = filename
                dst = os.path.join(backup_dir, "%s.backup" % basename)
                if action == 'backup':
                    bkp_set = [[src, dst]]
                elif action == 'restore':
                    bkp_set = [[dst, src]]
            else:
                # We have to make 2 backups, one of the bad image, another one
                # of the good image
                src_bad = filename
                src_good = os.path.join(backup_dir, "%s.backup" % basename)
                hsh = utils_misc.generate_random_string(4)
                dst_bad = (os.path.join(backup_dir, "%s.bad.%s" %
                                        (basename, hsh)))
                dst_good = (os.path.join(backup_dir, "%s.good.%s" %
                                         (basename, hsh)))
                if action == 'backup':
                    bkp_set = [[src_bad, dst_bad], [src_good, dst_good]]
                elif action == 'restore':
                    bkp_set = [[src_good, src_bad]]
            return bkp_set

        backup_dir = params.get("backup_dir", "")
        if not os.path.isabs(backup_dir):
            backup_dir = os.path.join(root_dir, backup_dir)
        backup_set = get_backup_set(self.image_filename, backup_dir,
                                    action, good)
        if self.is_remote_image():
            backup_func = self.copy_data_remote
        elif params.get('image_raw_device') == 'yes':
            backup_func = self.copy_data_raw
        else:
            backup_func = self.copy_data_file

        if action == 'backup':
            backup_size = 0
            for src, dst in backup_set:
                if os.path.isfile(src):
                    backup_size += os.path.getsize(src)
                else:
                    # TODO: get the size of block/remote images
                    if self.size:
                        backup_size += int(
                            float(utils_numeric.normalize_data_size(
                                self.size, order_magnitude="B"))
                        )

            s = os.statvfs(backup_dir)
            image_dir_free_disk_size = s.f_bavail * s.f_bsize
            logging.info("backup image size: %d, available size: %d.",
                         backup_size, image_dir_free_disk_size)
            if not self.is_disk_size_enough(backup_size,
                                            image_dir_free_disk_size):
                return

        # backup secret file if presented
        if self.encryption_config.key_secret:
            bk_set = get_backup_set(self.encryption_config.key_secret.filename,
                                    secret_dir, action, good)
            for src, dst in bk_set:
                self.copy_data_file(src, dst)

        # backup external data file
        if self.data_file:
            self.data_file.backup_image(self.data_file.params, root_dir,
                                        action, good, skip_existing)

        for src, dst in backup_set:
            if action == 'backup' and skip_existing and os.path.exists(dst):
                logging.debug("Image backup %s already exists, skipping...",
                              dst)
                continue
            backup_func(src, dst)

    def rm_backup_image(self):
        """
        Remove backup image
        """
        # remove external data file backup
        if self.data_file:
            self.data_file.rm_backup_image()

        backup_dir = utils_misc.get_path(self.root_dir,
                                         self.params.get("backup_dir", ""))
        image_name = os.path.join(backup_dir, "%s.backup" %
                                  os.path.basename(self.image_filename))
        logging.debug("Removing image file %s as requested", image_name)
        if os.path.exists(image_name):
            os.unlink(image_name)
        else:
            logging.warning("Image file %s not found", image_name)

    def save_image(self, params, filename, root_dir=None):
        """
        Save images to a path for later debugging.

        :param params: Dictionary containing the test parameters.
        :param filename: new filename for saved images.
        :param root_dir: directory for saved images.

        """
        src = self.image_filename
        if root_dir is None:
            root_dir = os.path.dirname(src)
        backup_func = self.copy_data_file
        if self.is_remote_image():
            backup_func = self.copy_data_remote
        elif params.get('image_raw_device') == 'yes':
            backup_func = self.copy_data_raw

        backup_size = 0
        if os.path.isfile(src):
            backup_size = os.path.getsize(src)
        else:
            # TODO: get the size of block/remote images
            if self.size:
                backup_size += int(
                    float(utils_numeric.normalize_data_size(
                        self.size, order_magnitude="B"))
                )
        s = os.statvfs(root_dir)
        image_dir_free_disk_size = s.f_bavail * s.f_bsize
        logging.info("Checking disk size on %s.", root_dir)
        if not self.is_disk_size_enough(backup_size,
                                        image_dir_free_disk_size):
            return

        backup_func(src, utils_misc.get_path(root_dir, filename))

    @staticmethod
    def is_disk_size_enough(required, available):
        """Check if available disk size is enough for the data copy."""
        minimum_disk_free = 1.2 * required
        if available < minimum_disk_free:
            logging.error("Free space: %s MB", (available / 1048576.))
            logging.error("Backup size: %s MB", (required / 1048576.))
            logging.error("Minimum free space acceptable: %s MB",
                          (minimum_disk_free / 1048576.))
            logging.error("Available disk space is not enough. Skipping...")
            return False
        return True

    def copy_data_remote(self, src, dst):
        pass

    @staticmethod
    def copy_data_raw(src, dst):
        """Using dd for raw device."""
        if os.path.exists(src):
            process.system("dd if=%s of=%s bs=4k conv=sync" % (src, dst))
        else:
            logging.info("No source %s, skipping dd...", src)

    @staticmethod
    def copy_data_file(src, dst):
        """Copy for files."""
        if os.path.isfile(src):
            logging.debug("Copying %s -> %s", src, dst)
            _dst = dst + '.part'
            shutil.copy(src, _dst)
            os.rename(_dst, dst)
        else:
            logging.info("No source file %s, skipping copy...", src)

    @staticmethod
    def clone_image(params, vm_name, image_name, root_dir):
        """
        Clone master image to vm specific file.

        :param params: Dictionary containing the test parameters.
        :param vm_name: Vm name.
        :param image_name: Master image name.
        :param root_dir: Base directory for relative filenames.
        """
        if not params.get("image_name_%s_%s" % (image_name, vm_name)):
            m_image_name = params.get("image_name", "image")
            vm_image_name = params.get("image_name_%s" % vm_name, "%s_%s" % (m_image_name, vm_name))
            if params.get("clone_master", "yes") == "yes":
                image_params = params.object_params(image_name)
                image_params["image_name"] = vm_image_name

                master_image = params.get("master_image_name")
                if master_image:
                    image_format = params.get("image_format", "qcow2")
                    m_image_fn = "%s.%s" % (master_image, image_format)
                    m_image_fn = utils_misc.get_path(root_dir, m_image_fn)
                else:
                    m_image_fn = get_image_filename(params, root_dir)
                image_fn = get_image_filename(image_params, root_dir)
                force_clone = params.get("force_image_clone", "no")
                if not os.path.exists(image_fn) or force_clone == "yes":
                    logging.info("Clone master image for vms.")
                    process.run(params.get("image_clone_command") %
                                (m_image_fn, image_fn))
            params["image_name_%s" % vm_name] = vm_image_name
            params["image_name_%s_%s" % (image_name, vm_name)] = vm_image_name

    @staticmethod
    def rm_cloned_image(params, vm_name, image_name, root_dir):
        """
        Remove vm specific file.

        :param params: Dictionary containing the test parameters.
        :param vm_name: Vm name.
        :param image_name: Master image name.
        :param root_dir: Base directory for relative filenames.
        """
        if params.get("image_name_%s_%s" % (image_name, vm_name)):
            m_image_name = params.get("image_name", "image")
            vm_image_name = "%s_%s" % (m_image_name, vm_name)
            if params.get("clone_master", "yes") == "yes":
                image_params = params.object_params(image_name)
                image_params["image_name"] = vm_image_name

                image_fn = get_image_filename(image_params, root_dir)

                logging.debug("Removing vm specific image file %s", image_fn)
                if os.path.exists(image_fn):
                    process.run(params.get("image_remove_command") % (image_fn))
                else:
                    logging.debug("Image file %s not found", image_fn)


class Rawdev(object):

    """
    Base class for raw storage devices such as iscsi and local disks
    """

    def __init__(self, params, root_dir, tag):
        """
        Init the default value for image object.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param tag: Image tag defined in parameter images
        """
        host_set_flag = params.get("host_setup_flag")
        if host_set_flag is not None:
            self.exec_cleanup = int(host_set_flag) & 2 == 2
        else:
            self.exec_cleanup = False
        if params.get("force_cleanup") == "yes":
            self.exec_cleanup = True
        self.image_name = tag


class Iscsidev(Rawdev):

    """
    Class for handle iscsi devices for VM
    """

    def __init__(self, params, root_dir, tag):
        """
        Init the default value for image object.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param tag: Image tag defined in parameter images
        """
        Rawdev.__init__(self, params, root_dir, tag)
        self.emulated_file_remove = False
        self.emulated_image = params.get("emulated_image")
        if self.emulated_image:
            self.emulated_image = os.path.join(root_dir, self.emulated_image)
            if params.get("emulated_file_remove", "no") == "yes":
                self.emulated_file_remove = True
        params["iscsi_thread_id"] = self.image_name
        self.iscsidevice = iscsi.Iscsi.create_iSCSI(params, root_dir=root_dir)
        self.device_id = params.get("device_id")
        self.iscsi_init_timeout = int(params.get("iscsi_init_timeout", 10))


class LVMdev(Rawdev):

    """
    Class for handle LVM devices for VM
    """

    def __init__(self, params, root_dir, tag):
        """
        Init the default value for image object.

        :param params: Dictionary containing the test parameters.
        :param root_dir: Base directory for relative filenames.
        :param tag: Image tag defined in parameter images
        """
        super(LVMdev, self).__init__(params, root_dir, tag)
        if params.get("emulational_device", "yes") == "yes":
            self.lvmdevice = lvm.EmulatedLVM(params, root_dir=root_dir)
        else:
            self.lvmdevice = lvm.LVM(params)



#  Copyright(c) 2013 Intel Corporation.
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms and conditions of the GNU General Public License,
#  version 2, as published by the Free Software Foundation.
#
#  This program is distributed in the hope it will be useful, but WITHOUT
#  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#
#  The full GNU General Public License is included in this distribution in
#  the file called "COPYING".

import os
import re
import logging
from tempfile import mktemp

from avocado.utils import process


_COMMAND_TABLE_DOC = """

Taken from http://fedoraproject.org/wiki/SysVinit_to_Systemd_Cheatsheet

service frobozz start
systemctl start frobozz.service
 Used to start a service (not reboot persistent)

service frobozz stop
systemctl stop frobozz.service
 Used to stop a service (not reboot persistent)

service frobozz restart
systemctl restart frobozz.service
 Used to stop and then start a service

service frobozz reload
systemctl reload frobozz.service
 When supported, reloads the config file without interrupting pending operations.

service frobozz condrestart
systemctl condrestart frobozz.service
 Restarts if the service is already running.

service frobozz status
systemctl status frobozz.service
 Tells whether a service is currently running.

ls /etc/rc.d/init.d/
systemctl list-unit-files --type=service (preferred)
 Used to list the services that can be started or stopped
ls /lib/systemd/system/*.service /etc/systemd/system/*.service
 Used to list all the services and other units

chkconfig frobozz on
systemctl enable frobozz.service
 Turn the service on, for start at next boot, or other trigger.

chkconfig frobozz off
systemctl disable frobozz.service
 Turn the service off for the next reboot, or any other trigger.

chkconfig frobozz
systemctl is-enabled frobozz.service
 Used to check whether a service is configured to start or not in the current environment.

chkconfig --list
systemctl list-unit-files --type=service(preferred)
ls /etc/systemd/system/*.wants/
 Print a table of services that lists which runlevels each is configured on or off

chkconfig frobozz --list
ls /etc/systemd/system/*.wants/frobozz.service
 Used to list what levels this service is configured on or off

chkconfig frobozz --add
systemctl daemon-reload
 Used when you create a new service file or modify any configuration


"""


def sysvinit_status_parser(cmdResult=None):
    """
    Parse method for service sub-command status.

    :return True  : if status is running or active.
    :return False : if status is stopped.
    :return None  : if status is unrecognized.
    """
    # If service is stopped, exit_status is also not zero.
    # So, we can't use exit_status to check result.
    result = cmdResult.stdout_text.lower()
    if re.search(r"unrecognized", result):
        return None
    dead_flags = [r"stopped", r"not running", r"dead"]
    for flag in dead_flags:
        if re.search(flag, result):
            return False
    # Is it running ?
    return bool(re.search(r"running", result))


def systemd_status_parser(cmdResult=None):
    """
    Parse method for service sub-command status.

    :return True  : if status is active(running).
    :return False : if status is stopped.
    :return None  : if status is un-loaded.
    """
    # If service is stopped, exit_status is also not zero.
    # So, we can't use exit_status to check result.
    result = cmdResult.stdout_text
    # check for systemctl status XXX.service.
    if not re.search(r"Loaded: loaded", result):
        return None
    # Is it active ?
    return (result.count("Active: active") > 0)


def sysvinit_list_parser(cmdResult=None):
    """
    Parse method for service sub-command list.

    :return in form of dict-like, including service name, status and so on

    For example::

        {"sshd":   {0: 'off', 1: 'off', 2: 'off', ..., 6: 'off'},
         "vsftpd": {0: 'off', 1: 'off', 2: 'off', ..., 6: 'off'},
         "xinetd": {'discard-dgram:': 'off', 'rsync:': 'on',...},
         ...
        }
    """
    if cmdResult.exit_status:
        raise process.CmdError(cmdResult.command, cmdResult)
    # The final dict to return.
    _service2statusOnTarget_dict = {}
    # Dict to store status on every target for each service.
    _status_on_target = {}
    # Dict to store the status for service based on xinetd.
    _service2statusOnXinet_dict = {}
    lines = cmdResult.stdout_text.strip().splitlines()
    for line in lines:
        sublines = line.strip().split()
        if len(sublines) == 8:
            # Service and status on each target.
            service_name = sublines[0]
            # Store the status of each target in _status_on_target.
            for target in range(7):
                status = sublines[target + 1].split(":")[-1]
                _status_on_target[target] = status
            _service2statusOnTarget_dict[
                service_name] = _status_on_target.copy()

        elif len(sublines) == 2:
            # Service based on xinetd.
            service_name = sublines[0].strip(":")
            status = sublines[-1]
            _service2statusOnXinet_dict[service_name] = status

        else:
            # Header or some lines useless.
            continue
    # Add xinetd based service in the main dict.
    _service2statusOnTarget_dict["xinetd"] = _service2statusOnXinet_dict
    return _service2statusOnTarget_dict


def systemd_list_parser(cmdResult=None):
    """
    Parse method for service sub-command list.

    :return in form of dict-like, including service name, status and so on

    For example::

        {"sshd": "enabled",
         "vsftpd": "disabled",
         "systemd-sysctl": "static",
         ...
        }
    """
    if cmdResult.exit_status:
        raise process.CmdError(cmdResult.command, cmdResult)
    # store service name and status.
    _service2status_dict = {}
    lines = cmdResult.stdout_text.strip().splitlines()
    for line in lines:
        sublines = line.strip().split()
        if (not len(sublines) == 2) or (not sublines[0].endswith("service")):
            # Some lines useless.
            continue
        service_name = sublines[0].rstrip(".service")
        status = sublines[-1]
        _service2status_dict[service_name] = status
    return _service2status_dict


# The parser shared by SysVInit and Systemd
def raw_status_parser(cmdResult=None):
    """
    Just return the result of service sub-command.
    """
    return cmdResult


# Please add new parser function into this dict
# when it's implemented.
# Format : {'sub-command' : 'function_parser'}
_SysVInit_parser_dict = {'status': sysvinit_status_parser,
                         'raw_status': raw_status_parser,
                         'list': sysvinit_list_parser}

_Systemd_parser_dict = {'status': systemd_status_parser,
                        'raw_status': raw_status_parser,
                        'list': systemd_list_parser}


def sysvinit_result_parser(command):
    """
    Parse results for sys_v style commands.

    :param command: service sub-command(string).
    :return: depends on sub-command.
    """
    if command in list(_SysVInit_parser_dict.keys()):
        return _SysVInit_parser_dict.get(command)
    else:
        return _ServiceResultParser.default_method


def systemd_result_parser(command):
    """
    Parse results for systemd style commands.

    :param command: service sub-command(string).
    :return: depends on sub-command.
    """
    if command in list(_Systemd_parser_dict.keys()):
        return _Systemd_parser_dict.get(command)
    else:
        return _ServiceResultParser.default_method


def sysvinit_command_generator(command):
    """
    Generate lists of command arguments for sys_v style inits.

    :param command: start,stop,restart, etc.
    :type command: str
    :return: list of commands to pass to process.run or similar function
    :rtype: builtin.list
    """
    command_name = "service"
    if command == "is_enabled":
        command_name = "chkconfig"
        command = ""
    elif command == 'enable':
        command_name = "chkconfig"
        command = "on"
    elif command == 'disable':
        command_name = "chkconfig"
        command = "off"
    elif command == "raw_status":
        command = "status"
    elif command == 'list':
        # noinspection PyUnusedLocal
        def list_command(service_name):
            return ["chkconfig", "--list"]
        return list_command
    elif command == "set_target":
        def set_target_command(target):
            target = convert_systemd_target_to_runlevel(target)
            return ["telinit", target]
        return set_target_command
    # Do not need reset failed in sys_v style system.
    elif command == "reset_failed":
        def true_command(service_name):
            return ["true"]
        return true_command

    def method(service_name):
        return [command_name, service_name, command]
    return method


def systemd_command_generator(command):
    """
    Generate list of command line argument strings for systemctl.
    One argument per string for compatibility Popen

    WARNING: If systemctl detects that it is running on a tty it will use color,
    pipe to $PAGER, change column sizes and not truncate unit names.
    Use --no-pager to suppress pager output, or set PAGER=cat in the environment.
    You may need to take other steps to suppress color output.
    See https://bugzilla.redhat.com/show_bug.cgi?id=713567

    :param command: start,stop,restart, etc.
    :type command: str
    :return: list of command and arguments to pass to process.run or similar functions
    :rtype: builtin.list
    """
    command_name = "systemctl"
    if command == "is_enabled":
        command = "is-enabled"
    elif command == "list":
        # noinspection PyUnusedLocal
        def list_command(service_name):
            # systemctl pipes to `less` or $PAGER by default. Workaround this
            # add '--full' to avoid systemctl truncates service names.
            return [command_name, "list-unit-files",
                    "--type=service", "--no-pager", "--full"]
        return list_command
    elif command == "set_target":
        def set_target_command(target):
            return [command_name, "isolate", target]
        return set_target_command
    elif command == "reset_failed":
        command = "reset-failed"
    elif command == "raw_status":
        command = "status"

    def method(service_name):
        return [command_name, command, service_name]
    return method


COMMANDS = (
    "start",
    "stop",
    "reload",
    "restart",
    "condrestart",
    "status",
    "enable",
    "disable",
    "is_enabled",
    "list",
    "set_target",
    "reset_failed",
    "raw_status"
)


class _ServiceCommTool(object):

    """
    Provide an interface to complete the similar initialization
    """

    def __init__(self, func, command_list):
        """
        common __init__ function used to create staticmethod

        :param func:         function name
        :param command_list: list of all the commands, e.g. start, stop, etc
        """
        self.commands = command_list
        for command in self.commands:
            setattr(self, command, func(command))


class _ServiceResultParser(_ServiceCommTool):

    """
    A class that contains staticmethods to parse the result of service command.
    """

    def __init__(self, result_parser, command_list=COMMANDS):
        """
        Create staticmethods for each command in command_list[],
        which used to parse result of command

        :param result_parser: function used to parse result of command.
        :type  result_parser: function
        """
        super(_ServiceResultParser, self).__init__(func=result_parser,
                                                   command_list=command_list)

    @staticmethod
    def default_method(cmdResult):
        """
        Parse result for the command, which neithor 'list' nor 'status'.

        return True if command was executed successfully.
        """
        if cmdResult.exit_status:
            logging.debug(cmdResult)
            return False
        else:
            return True


class _ServiceCommandGenerator(_ServiceCommTool):

    """
    A class that contains staticmethods that generate partial functions that
    generate command lists for starting/stopping services.
    """

    def __init__(self, command_generator, command_list=COMMANDS):
        """
        Create staticmethods for each command in command_list[]

        :param command_generator: function used to generate command string
        :type  command_generator: function
        """
        super(_ServiceCommandGenerator, self).__init__(func=command_generator,
                                                       command_list=command_list)


class _SpecificServiceManager(object):

    def __init__(self, service_name, service_command_generator, service_result_parser, run=process.run):
        """
        Create staticmethods that call process.run with the given service_name
        for each command in service_command_generator.

        lldpad = SpecificServiceManager("lldpad",
                                        auto_create_specific_service_command_generator())
        lldpad.start()
        lldpad.stop()

        :param service_name: init service name or systemd unit name
        :type service_name: str
        :param service_command_generator: a sys_v_init or systemd command generator
        :type service_command_generator: _ServiceCommandGenerator
        :param run: function that executes the commands and return CmdResult object, default process.run
        :type run: function
        """
        for cmd in service_command_generator.commands:
            setattr(self, cmd,
                    self.generate_run_function(run,
                                               getattr(service_result_parser, cmd),
                                               getattr(service_command_generator, cmd),
                                               service_name))

    @staticmethod
    def generate_run_function(run_func, parse_func, command, service_name):
        """
        Generate the wrapped call to process.run for the given service_name.

        :param run_func:  function to execute command and return CmdResult object.
        :type run_func:  function
        :param parse_func: function to parse the result from run.
        :type parse_func: function
        :param command: partial function that generates the command list
        :type command: function
        :param service_name: init service name or systemd unit name
        :type service_name: str
        :return: wrapped process.run function.
        :rtype: function
        """
        def run(**kwargs):
            """
            Wrapped process.run invocation that will start, stop, restart, etc. a service.

            :param kwargs: extra arguments to process.run, .e.g. timeout. But not for ignore_status.
                           We need a CmdResult to parse and raise a exceptions.TestError if command failed.
                           We will not let the CmdError out.
            :return: result of parse_func.
            """
            logging.debug("Setting ignore_status to True.")
            kwargs["ignore_status"] = True
            result = run_func(" ".join(command(service_name)), **kwargs)
            result.stdout = result.stdout_text
            result.stderr = result.stderr_text
            return parse_func(result)
        return run


class _GenericServiceManager(object):

    """
    Base class for SysVInitServiceManager and SystemdServiceManager.
    """

    def __init__(self, service_command_generator, service_result_parser, run=process.run):
        """
        Create staticmethods for each service command, e.g. start, stop, restart.
        These staticmethods take as an argument the service to be started or stopped.

        systemd = SpecificServiceManager(auto_create_specific_service_command_generator())
        systemd.start("lldpad")
        systemd.stop("lldpad")

        :param service_command_generator: a sys_v_init or systemd command generator
        :type service_command_generator: _ServiceCommandGenerator
        :param run: function to call the run the commands, default process.run
        :type run: function
        """
        # create staticmethods in class attributes (not used)
        # for cmd in service_command_generator.commands:
        #     setattr(self.__class__, cmd,
        #             staticmethod(self.generate_run_function(run, getattr(service_command_generator, cmd))))
        # create functions in instance attributes
        for cmd in service_command_generator.commands:
            setattr(self, cmd,
                    self.generate_run_function(run,
                                               getattr(service_result_parser, cmd),
                                               getattr(service_command_generator, cmd)))

    @staticmethod
    def generate_run_function(run_func, parse_func, command):
        """
        Generate the wrapped call to process.run for the service command, "service" or "systemctl"

        :param run_func:  process.run
        :type run_func:  function
        :param command: partial function that generates the command list
        :type command: function
        :return: wrapped process.run function.
        :rtype: function
        """
        def run(service="", **kwargs):
            """
            Wrapped process.run invocation that will start, stop, restart, etc. a service.

            :param service: service name, e.g. crond, dbus, etc.
            :param kwargs: extra arguments to process.run, .e.g. timeout. But not for ignore_status.
                           We need a CmdResult to parse and raise a exceptions.TestError if command failed.
                           We will not let the CmdError out.
            :return: result of parse_func.
            """
            logging.debug("Setting ignore_status to True.")
            kwargs["ignore_status"] = True
            result = run_func(" ".join(command(service)), **kwargs)
            return parse_func(result)
        return run


class _SysVInitServiceManager(_GenericServiceManager):

    """
    Concrete class that implements the SysVInitServiceManager
    """

    def __init__(self, service_command_generator, service_result_parser, run=process.run):
        """
        Create the GenericServiceManager for SysV services.

        :param service_command_generator:
        :type service_command_generator: _ServiceCommandGenerator
        :param run: function to call to run the commands, default process.run
        :type run: function
        """
        super(
            _SysVInitServiceManager, self).__init__(service_command_generator,
                                                    service_result_parser, run)

    # @staticmethod
    # def change_default_runlevel(runlevel='3'):
    #     """
    #     Set the default sys_v runlevel
    #
    #     :param runlevel: sys_v runlevel to set as default in inittab
    #     :type runlevel: str
    #     """
    #     raise NotImplemented


def convert_sysv_runlevel(level):
    """
    Convert runlevel to systemd target.

    :param level: sys_v runlevel
    :type level: str or int
    :return: systemd target
    :rtype: str
    :raise ValueError: when runlevel is unknown
    """
    runlevel = str(level)
    if runlevel == '0':
        target = "poweroff.target"
    elif runlevel in ['1', "s", "single"]:
        target = "rescue.target"
    elif runlevel in ['2', '3', '4']:
        target = "multi-user.target"
    elif runlevel == '5':
        target = "graphical.target"
    elif runlevel == '6':
        target = "reboot.target"
    else:
        raise ValueError("unknown runlevel %s" % level)
    return target


def convert_systemd_target_to_runlevel(target):
    """
    Convert systemd target to runlevel.

    :param target: systemd target
    :type target: str
    :return: sys_v runlevel
    :rtype: str
    :raise ValueError: when systemd target is unknown
    """
    if target == "poweroff.target":
        runlevel = '0'
    elif target == "rescue.target":
        runlevel = 's'
    elif target == "multi-user.target":
        runlevel = '3'
    elif target == "graphical.target":
        runlevel = '5'
    elif target == "reboot.target":
        runlevel = '6'
    else:
        raise ValueError("unknown target %s" % target)
    return runlevel


class _SystemdServiceManager(_GenericServiceManager):

    """
    Concrete class that implements the SystemdServiceManager
    """

    def __init__(self, service_command_generator, service_result_parser, run=process.run):
        """
        Create the GenericServiceManager for systemd services.

        :param service_command_generator:
        :type service_command_generator: _ServiceCommandGenerator
        :param run: function to call to run the commands, default process.run
        :type run: function
        """
        super(_SystemdServiceManager, self).__init__(service_command_generator,
                                                     service_result_parser, run)

    @staticmethod
    def change_default_runlevel(runlevel='multi-user.target'):
        # atomic symlinking, symlink and then rename
        """
        Set the default systemd target.
        Create the symlink in a temp directory and then use
        atomic rename to move the symlink into place.

        :param runlevel: default systemd target
        :type runlevel: str
        """
        tmp_symlink = mktemp(dir="/etc/systemd/system")
        os.symlink("/usr/lib/systemd/system/%s" % runlevel, tmp_symlink)
        os.rename(tmp_symlink, "/etc/systemd/system/default.target")


class Factory(object):

    """
    Class to create different kinds of ServiceManager.
    The all interfaces to create manager are staticmethod,
    so we do not have to create an instance of factory
    when create manager.

    * GenericServiceManager:
        * Interface: create_generic_service()
        * Description: Object to manage the all services(lldp, sshd and so on).
                You can list the all services by GenericServiceManager.list().
                And you can operate any service by passing the service name,
                such as GenericServiceManager.start("sshd").

                Example:
                # Get the system service manager
                service_manager = Factory.create_generic_service()

                # Stating service/unit "sshd"
                service_manager.start("sshd")

                # Getting a list of available units
                units = service_manager.list()

    * SpecificServiceManager:
        * interface: create_specific_service(service_name)
        * description: Object to manage specific service(such as sshd). You can
                not operate the other services nor list the all information on
                this host.

                # Get the specific service manager for sshd
                sshd = Factory.create_specific_service("sshd")
                sshd.start()
                sshd.stop()

    After all, there is an unified interface to create both of them,
    create_service(service_name=None).

    If we pass a service_name to it, it will return a SpecificServiceManager,
    otherwise, it will return GenericServiceManager.
    """

    class FactoryHelper(object):

        """
        Internal class to help create service manager.

        Provide some functions to auto detect system type.
        And auto create command_generator and result_parser.
        """
        _command_generators = {"init": sysvinit_command_generator,
                               "systemd": systemd_command_generator}

        _result_parsers = {"init": sysvinit_result_parser,
                           "systemd": systemd_result_parser}

        _service_managers = {"init": _SysVInitServiceManager,
                             "systemd": _SystemdServiceManager}

        def __init__(self, run=process.run):
            """
            Init a helper to create service manager.

            :param run: Funtion to run command.
            :type: process.run-like function.
            """
            result = run("true")
            if not isinstance(result, process.CmdResult):
                raise ValueError("Param run is a/an %s, "
                                 "but not an instance of process.CmdResult."
                                 % (type(result)))
            self.run = run
            self.init_name = self.get_name_of_init()

        def get_name_of_init(self):
            """
            Internal function to determine what executable is PID 1,
            :return: executable name for PID 1, aka init
            :rtype:  str
            """
            output = self.run("ps -o comm 1").stdout_text
            return output.splitlines()[-1].strip()

        def get_generic_service_manager_type(self):
            """
            Get the ServiceManager type using the auto-detect init command.

            :return: Subclass type of _GenericServiceManager from the current init command.
            :rtype: _SysVInitServiceManager or _SystemdServiceManager.
            """
            return self._service_managers[self.init_name]

        def get_generic_service_result_parser(self):
            """
            Get the ServiceResultParser using the auto-detect init command.

            :return: ServiceResultParser fro the current init command.
            :rtype: _ServiceResultParser
            """
            result_parser = self._result_parsers[self.init_name]
            return _ServiceResultParser(result_parser)

        def get_generic_service_command_generator(self):
            """
            Lazy initializer for ServiceCommandGenerator using the auto-detect init command.

            :return: ServiceCommandGenerator for the current init command.
            :rtype: _ServiceCommandGenerator
            """
            command_generator = self._command_generators[self.init_name]
            return _ServiceCommandGenerator(command_generator)

        def get_specific_service_result_parser(self):
            """
            Create a class that will create partial functions that generate result_parser
            for the current init command.

            :return: A ServiceResultParser for the auto-detected init command.
            :rtype: _ServiceResultParser
            """
            result_parser = self._result_parsers[self.init_name]
            # remove list method
            command_list = [
                c for c in COMMANDS if c not in ["list", "set_target"]]
            return _ServiceResultParser(result_parser, command_list)

        def get_specific_service_command_generator(self):
            """
            Create a class that will create partial functions that generate commands
            for the current init command.

            ::

                lldpad = SpecificServiceManager("lldpad",
                             auto_create_specific_service_command_generator())
                lldpad.start()
                lldpad.stop()

            :return: A ServiceCommandGenerator for the auto-detected init command.
            :rtype: _ServiceCommandGenerator
            """
            command_generator = self._command_generators[self.init_name]
            # remove list method
            command_list = [
                c for c in COMMANDS if c not in ["list", "set_target"]]
            return _ServiceCommandGenerator(command_generator, command_list)

    @staticmethod
    def create_generic_service(run=process.run):
        """
        Detect which init program is being used, init or systemd and return a
        class with methods to start/stop services.

        ::

            # Get the system service manager
            service_manager = Factory.create_generic_service()

            # Stating service/unit "sshd"
            service_manager.start("sshd")

            # Getting a list of available units
            units = service_manager.list()

            # Disabling and stopping a list of services
            services_to_disable = ['ntpd', 'httpd']
            for s in services_to_disable:
                service_manager.disable(s)
                service_manager.stop(s)

        :return: SysVInitServiceManager or SystemdServiceManager
        :rtype: _GenericServiceManager
        """
        helper = Factory.FactoryHelper(run)
        service_manager = helper.get_generic_service_manager_type()
        command_generator = helper.get_generic_service_command_generator()
        result_parser = helper.get_generic_service_result_parser()
        return service_manager(command_generator, result_parser, run)

    @staticmethod
    def create_specific_service(service_name, run=process.run):
        """

        # Get the specific service manager for sshd
        sshd = Factory.create_specific_service("sshd")
        sshd.start()
        sshd.stop()
        sshd.reload()
        sshd.restart()
        sshd.condrestart()
        sshd.status()
        sshd.enable()
        sshd.disable()
        sshd.is_enabled()

        :param service_name: systemd unit or init.d service to manager
        :type service_name: str
        :return: SpecificServiceManager that has start/stop methods
        :rtype: _SpecificServiceManager
        """
        helper = Factory.FactoryHelper(run)
        command_generator = helper.get_specific_service_command_generator()
        result_parser = helper.get_specific_service_result_parser()

        return _SpecificServiceManager(service_name,
                                       command_generator,
                                       result_parser,
                                       run)

    @staticmethod
    def create_service(service_name=None, run=process.run):
        """
        # Unified interface for generic and specific service manager.

        :return: _SpecificServiceManager if service_name is not None,
                _GenericServiceManager if service_name is None.
        """
        if service_name:
            return Factory.create_specific_service(service_name, run)
        return Factory.create_generic_service(run)


def singleton_class(class_name):
    services = {}
    class_instance = {}

    def get_class(*args, **kwargs):
        host = "local"
        if kwargs.get("session"):
            host = "remote_%s" % kwargs["session"]
        if host not in services or args[-1] not in services[host]:
            class_object = class_name(*args, **kwargs)
            class_instance[args[-1]] = class_object
            services[host] = class_instance
        return services[host][args[-1]]
    return get_class


@singleton_class
class Service(object):
    """
    class to handle service operations in Remote machine/Guest
    """
    def __init__(self, service_name, session=None, timeout=120):
        """
        Initialize the service return Factory object if session is None

        :param service_name: name of service to be operated
        :param session: ShellSession Object of remote/guest machine
        :param timeout: ShellSession timeout
        """
        self.session = session
        if not self.session:
            self.service = Factory.create_service(service_name)
        self.service_name = service_name
        self.timeout = timeout

    def command(self, operate, output=False):
        """
        Method to trigger the systemctl commands to operate on the service

        :param operate: operation name start/stop/status
        :param output: True to return output else return status
        """
        cmd = "systemctl %s %s" % (operate, self.service_name)
        status, output = self.session.cmd_status_output(cmd, timeout=self.timeout)
        if status != 0:
            logging.error("%s returned unexpected status %s", cmd, status)
        if output:
            return output
        return status

    def restart(self):
        """
        Method to perform restart operation on the given service
        """
        if not self.session:
            return self.service.restart()
        return self.command("restart") == 0

    def start(self):
        """
        Method to perform start operation on the given service
        """
        if not self.session:
            return self.service.start()
        return self.command("start") == 0

    def stop(self):
        """
        Method to perform stop operation on the given service
        """
        if not self.session:
            return self.service.stop()
        return self.command("stop") == 0

    def status(self, regex=r"active \(running\)"):
        """
        Method to get the status of given service
        """
        if not self.session:
            return self.service.status()
        output = self.command("status", output=True)
        return bool(re.search(regex, output))



import sys
import threading
from traceback import format_exception

# Add names you want to be imported by 'from errors import *' to this list.
# This must be list not a tuple as we modify it to include all of our
# the Exception classes we define below at the end of this file.
__all__ = ['format_error', 'context_aware', 'context', 'get_context',
           'exception_context']


def format_error():
    t, o, tb = sys.exc_info()
    trace = format_exception(t, o, tb)
    # Clear the backtrace to prevent a circular reference
    # in the heap -- as per tutorial
    tb = ''

    return ''.join(trace)


# Exception context information:
# ------------------------------
# Every function can have some context string associated with it.
# The context string can be changed by calling context(str) and cleared by
# calling context() with no parameters.
# get_context() joins the current context strings of all functions in the
# provided traceback.  The result is a brief description of what the test was
# doing in the provided traceback (which should be the traceback of a caught
# exception).
#
# For example: assume a() calls b() and b() calls c().
#
# @error.context_aware
# def a():
#     error.context("hello")
#     b()
#     error.context("world")
#     error.get_context() ----> 'world'
#
# @error.context_aware
# def b():
#     error.context("foo")
#     c()
#
# @error.context_aware
# def c():
#     error.context("bar")
#     error.get_context() ----> 'hello --> foo --> bar'
#
# The current context is automatically inserted into exceptions raised in
# context_aware functions, so usually test code doesn't need to call
# error.get_context().

ctx = threading.local()


def _new_context(s=""):
    if not hasattr(ctx, "contexts"):
        ctx.contexts = []
    ctx.contexts.append(s)


def _pop_context():
    ctx.contexts.pop()


def context(s="", log=None):
    """
    Set the context for the currently executing function and optionally log it.

    :param s: A string.  If not provided, the context for the current function
            will be cleared.
    :param log: A logging function to pass the context message to.  If None, no
            function will be called.
    """
    ctx.contexts[-1] = s
    if s and log:
        log("Context: %s" % get_context())


def base_context(s="", log=None):
    """
    Set the base context for the currently executing function and optionally
    log it.  The base context is just another context level that is hidden by
    default.  Functions that require a single context level should not use
    base_context().

    :param s: A string.  If not provided, the base context for the current
            function will be cleared.
    :param log: A logging function to pass the context message to.  If None, no
            function will be called.
    """
    ctx.contexts[-1] = ""
    ctx.contexts[-2] = s
    if s and log:
        log("Context: %s" % get_context())


def get_context():
    """Return the current context (or None if none is defined)."""
    if hasattr(ctx, "contexts"):
        return " --> ".join([s for s in ctx.contexts if s])


def exception_context(e):
    """Return the context of a given exception (or None if none is defined)."""
    if hasattr(e, "_context"):
        return e._context


def set_exception_context(e, s):
    """Set the context of a given exception."""
    e._context = s


def join_contexts(s1, s2):
    """Join two context strings."""
    if s1:
        if s2:
            return "%s --> %s" % (s1, s2)
        else:
            return s1
    else:
        return s2


def context_aware(fn):
    """A decorator that must be applied to functions that call context()."""
    def new_fn(*args, **kwargs):
        _new_context()
        _new_context("(%s)" % fn.__name__)
        try:
            try:
                return fn(*args, **kwargs)
            except Exception as e:
                if not exception_context(e):
                    set_exception_context(e, get_context())
                raise
        finally:
            _pop_context()
            _pop_context()
    new_fn.__name__ = fn.__name__
    new_fn.__doc__ = fn.__doc__
    new_fn.__dict__.update(fn.__dict__)
    return new_fn


def _context_message(e):
    s = exception_context(e)
    if s:
        return "    [context: %s]" % s
    else:
        return ""



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2015
# Author: Lucas Meneghel Rodrigues <lmr@redhat.com>

"""
Avocado plugin that augments 'avocado list' with avocado-virt related options.
"""

import os
import sys

from avocado.core.loader import loader
from avocado.core.plugin_interfaces import CLI

from virttest.compat import get_settings_value
from .vt import add_basic_vt_options, add_qemu_bin_vt_option
from ..loader import VirtTestLoader


# The original virt-test runner supports using autotest from a git checkout,
# so we'll have to support that as well. The code below will pick up the
# environment variable $AUTOTEST_PATH and do the import magic needed to make
# the autotest library available in the system.
AUTOTEST_PATH = None

if 'AUTOTEST_PATH' in os.environ:
    AUTOTEST_PATH = os.path.expanduser(os.environ['AUTOTEST_PATH'])
    CLIENT_DIR = os.path.join(os.path.abspath(AUTOTEST_PATH), 'client')
    SETUP_MODULES_PATH = os.path.join(CLIENT_DIR, 'setup_modules.py')
    if not os.path.exists(SETUP_MODULES_PATH):
        raise EnvironmentError("Although AUTOTEST_PATH has been declared, "
                               "%s missing." % SETUP_MODULES_PATH)
    import imp
    SETUP_MODULES = imp.load_source('autotest_setup_modules',
                                    SETUP_MODULES_PATH)
    SETUP_MODULES.setup(base_path=CLIENT_DIR,
                        root_module_name="autotest.client")

# The code below is used by this plugin to find the virt test directory,
# so that it can load the virttest python lib, used by the plugin code.
# If the user doesn't provide the proper configuration, the plugin will
# fail to load.
VIRT_TEST_PATH = None

if 'VIRT_TEST_PATH' in os.environ:
    VIRT_TEST_PATH = os.environ['VIRT_TEST_PATH']
else:
    VIRT_TEST_PATH = get_settings_value(section='virt_test',
                                        key='virt_test_path', default=None)

if VIRT_TEST_PATH is not None:
    sys.path.append(os.path.expanduser(VIRT_TEST_PATH))

from virttest import data_dir   # pylint: disable=C0413


_PROVIDERS_DOWNLOAD_DIR = os.path.join(data_dir.get_test_providers_dir(),
                                       'downloads')

try:
    assert len(os.listdir(_PROVIDERS_DOWNLOAD_DIR)) != 0
except (OSError, AssertionError):
    raise EnvironmentError("Bootstrap missing. "
                           "Execute 'avocado vt-bootstrap' or disable this "
                           "plugin to get rid of this message")


class VTLister(CLI):

    """
    Avocado VT - implements legacy virt-test listing
    """

    name = 'vt-list'
    description = "Avocado-VT/virt-test support for 'list' command"

    def configure(self, parser):
        """
        Add the subparser for the run action.

        :param parser: Main test runner parser.
        """
        list_subcommand_parser = parser.subcommands.choices.get('list', None)
        if list_subcommand_parser is None:
            return

        vt_compat_group_lister = list_subcommand_parser.add_argument_group(
            'Virt-Test compat layer - Lister options')
        vt_compat_group_lister.add_argument("--vt-list-guests",
                                            action="store_true",
                                            default=False,
                                            help="Also list the available "
                                            "guests (this option ignores the "
                                            "--vt-config and --vt-guest-os)")
        vt_compat_group_lister.add_argument("--vt-list-archs", default=False,
                                            action="store_true",
                                            help="Also list the available "
                                            "arch/machines for the given "
                                            "guest OS. (Use \"--vt-guest-os "
                                            "''\" to see all combinations; "
                                            "--vt-config --vt-machine-type "
                                            "and --vt-arch args are ignored)")
        add_basic_vt_options(vt_compat_group_lister)
        add_qemu_bin_vt_option(vt_compat_group_lister)

    def run(self, config):
        loader.register_plugin(VirtTestLoader)



"""
Utility classes and functions to handle KVM Qtree parsing and verification.

:author: Lukas Doktor <ldoktor@redhat.com>
:copyright: 2012 Red Hat Inc.
"""
import json
import logging
import os
import re

from six.moves import xrange

from . import storage
from . import data_dir
from . import utils_misc
from . import arch

import six


OFFSET_PER_LEVEL = 2

_RE_BLANKS = re.compile(r'^([ ]*)')
_RE_CLASS = re.compile(r'^class ([^,]*), addr (\w\w:\w\w.\w+), pci id '
                       '(\w{4}:\w{4}) \(sub (\w{4}:\w{4})\)')


class IncompatibleTypeError(TypeError):

    def __init__(self, prop, desired_type, value):
        TypeError.__init__(self)
        self.prop = prop
        self.desired = desired_type
        self.value = value

    def __str__(self):
        return "%s have to be %s, not %s" % (self.prop, type(self.desired),
                                             type(self.value))


class QtreeNode(object):

    """
    Generic Qtree node
    """

    def __init__(self):
        self.parent = None      # Parent node
        self.qtree = {}         # List of qtree attributes
        self.children = []     # List of child nodes
        self.params = {}        # generated params from qtree

    def __str__(self):
        out = self.str_short()
        if self.parent:
            out += "\n[parent]\n %s" % self.parent.str_short()
        if self.qtree:
            out += "\n[info qtree]"
        for tmp in six.iteritems(self.qtree):
            out += "\n %s = %s" % (tmp[0], tmp[1])
        if self.children:
            out += "\n[children]"
        for tmp in self.children:
            out += "\n %s" % tmp.str_short()
        if self.params:
            out += "\n[params]"
        for tmp in six.iteritems(self.params):
            out += "\n %s = %s" % (tmp[0], [tmp[1]])
        return out

    def set_parent(self, parent):
        if not isinstance(parent, QtreeNode) and parent is not None:
            raise IncompatibleTypeError('parent', QtreeNode(), parent)
        self.parent = parent

    def get_parent(self):
        return self.parent

    def add_child(self, child):
        if not isinstance(child, QtreeNode):
            raise IncompatibleTypeError('child', QtreeNode(), child)
        self.children.append(child)

    def replace_child(self, oldchild, newchild):
        if oldchild not in self.children:
            raise ValueError('child %s not in children %s' % (oldchild,
                                                              self.children))
        self.add_child(newchild)
        self.children.remove(oldchild)

    def get_children(self):
        return self.children

    def set_qtree(self, qtree):
        if not isinstance(qtree, dict):
            raise IncompatibleTypeError('qtree', {}, qtree)
        self.qtree = qtree

    def set_qtree_prop(self, prop, value):
        if prop in self.qtree:
            raise ValueError("Property %s = %s, not rewriting with %s" % (
                prop, self.qtree.get(prop), value))
        self.update_qtree_prop(prop, value)

    def update_qtree_prop(self, prop, value):
        if prop.startswith("bus-prop: "):
            prop = prop[10:]
        if prop.startswith("dev-prop: "):
            prop = prop[10:]
        self.qtree[prop] = value

    def get_qtree(self):
        return self.qtree

    def guess_type(self):
        """ Detect type of this object from qtree props """
        return QtreeNode

    def str_short(self):
        return "id: '%s', type: %s" % (self.qtree.get('id'), type(self))

    def str_qtree(self):
        out = "%s" % self.str_short()
        for child in self.children:
            for line in child.str_qtree().splitlines():
                out += "\n  %s" % line
        return out

    def generate_params(self):
        pass

    def get_params(self):
        return self.params

    def update_params(self, param, value):
        self.params[param] = value

    def verify(self):
        pass


class QtreeBus(QtreeNode):

    """ bus: qtree object """

    def __init__(self):
        super(QtreeBus, self).__init__()

    def add_child(self, child):
        if not isinstance(child, QtreeDev):
            raise IncompatibleTypeError('child', QtreeDev(), child)
        super(QtreeBus, self).add_child(child)

    def guess_type(self):
        return QtreeBus


class QtreeDev(QtreeNode):

    """ dev: qtree object """

    def __init__(self):
        super(QtreeDev, self).__init__()

    def add_child(self, child):
        if not isinstance(child, QtreeBus):
            raise IncompatibleTypeError('child', QtreeBus(), child)
        super(QtreeDev, self).add_child(child)

    def guess_type(self):
        guess = {True: QtreeDisk, False: QtreeDev}
        is_disk = ('drive' in self.qtree)
        # HOOK when usb-storage-containter is detected as disk
        non_disk_type = ('usb-storage', 'spapr-nvram', 'cfi.pflash01')
        is_disk = (is_disk and (self.qtree['type'] not in non_disk_type))
        return guess[is_disk]


class QtreeDisk(QtreeDev):

    """ qtree disk object """

    def __init__(self):
        super(QtreeDisk, self).__init__()
        self.block = {}     # Info from 'info block'

    def __str__(self):
        out = super(QtreeDisk, self).__str__()
        if self.block:
            out += "\n[info block]"
        for tmp in six.iteritems(self.block):
            out += "\n%s = %s" % (tmp[0], tmp[1])
        return out

    def set_block_prop(self, prop, value):
        if prop in self.block:
            raise ValueError("Property %s = %s, not rewriting with %s" % (
                prop, self.block.get(prop), value))
        self.update_block_prop(prop, value)

    def update_block_prop(self, prop, value):
        if prop.startswith("bus-prop: "):
            prop = prop[10:]
        if prop.startswith("dev-prop: "):
            prop = prop[10:]
        self.block[prop] = value

    def get_block(self):
        return self.block

    def generate_params(self):
        def parse_json(data):
            return json.loads(data[5:])

        def get_image_name(block_file):
            if block_file.startswith('json:'):
                return parse_json(block_file)['file']['filename']
            return os.path.realpath(block_file)

        if not self.qtree or not self.block:
            raise ValueError("Node doesn't have qtree or block info yet.")
        if self.block.get('backing_file'):
            self.params['image_snapshot'] = 'yes'
            self.params['image_name'] = get_image_name(self.block.get('backing_file'))
        elif self.block.get('file'):
            self.params['image_name'] = get_image_name(self.block.get('file'))
        else:
            raise ValueError("Missing 'file' or 'backing_file' information "
                             "in self.block.")
        if self.block.get('ro') and self.block.get('ro') != '0':
            self.params['image_readonly'] = 'yes'
        self.params['drive_format'] = self.qtree.get('type')

    def get_qname(self):
        return re.sub("['\"]", "", self.qtree.get('drive'))


class QtreeContainer(object):

    """ Container for Qtree """

    def __init__(self):
        self.nodes = None

    def get_qtree(self):
        """ :return: root of qtree """
        if self.nodes:
            return self.nodes[-1]

    def get_nodes(self):
        """
        :return: flat list of all qtree nodes (last one is main-system-bus)
        """
        return self.nodes

    def parse_info_qtree(self, info):
        """
        Parses 'info qtree' output. Creates list of self.nodes. Last node is
        the main-system-bus (whole qtree)
        """
        def _replace_node(old, newtype):
            if isinstance(old, newtype):
                return old
            new = newtype()
            new.set_parent(old.get_parent())
            new.get_parent().replace_child(old, new)
            new.set_qtree(old.get_qtree())
            for child in old.get_children():
                child.set_parent(new)
                new.add_child(child)
            return new

        def _hook_usb2_disk(node):
            """
            usb2 disk - from point of qtree - is scsi disk inside the
            usb-storage device.
            """
            # We're looking for scsi disk with grand-grand parent of
            # usb sorage type
            if not isinstance(node, QtreeDisk):
                return  # Not a disk
            if not node.get_qtree().get('type').startswith('scsi'):
                return  # Not scsi disk
            if not (node.get_parent() and node.get_parent().get_parent()):
                return  # Doesn't have grand-grand parent
            if not (node.get_parent().get_parent().get_qtree().get('type') ==
                    'usb-storage'):
                return  # grand-grand parent is not usb-storage
            # This disk is not scsi disk, it's virtual usb-storage drive
            node.update_qtree_prop('type', 'usb2')
        info = info.split('\n')
        current = None
        offset = 0
        self.nodes = []
        line = info.pop(0)
        while True:
            _offset = len(_RE_BLANKS.match(line).group(0))
            if not line.strip():
                if len(info) == 0:
                    break
                line = info.pop(0)
                continue
            if _offset >= offset:
                offset = _offset
                line = line[offset:]
                # Strip out all dev-prop/bus-prop/...
                # bus/dev/prop
                if line.startswith('bus: '):
                    # bus: scsi.0
                    new = QtreeBus()
                    if current:
                        current.add_child(new)
                        new.set_parent(current)
                    current = new
                    offset += OFFSET_PER_LEVEL
                    line = ['id', line[5:].strip()]
                elif line.startswith('dev: '):
                    # dev: scsi-disk, id ""
                    new = QtreeDev()
                    if current:
                        current.add_child(new)
                        new.set_parent(current)
                    current = new
                    line = line[5:].split(',')
                    line[1] = line[1].strip()
                    q_id = line[1][4:-1]
                    if len(q_id) > 0:
                        current.set_qtree_prop('id', q_id)
                    offset += OFFSET_PER_LEVEL
                    line = ['type', line[0]]
                elif _RE_CLASS.match(line):
                    # class IDE controller, addr 00:01.1, pci id 8086:7010 (..
                    line = _RE_CLASS.match(line).groups()
                    current.set_qtree_prop('class_addr', line[1])
                    current.set_qtree_prop('class_pciid', line[2])
                    current.set_qtree_prop('class_sub', line[3])
                    line = ['class_name', line[0]]
                elif '=' in line:
                    # bus-prop: addr = 02.0
                    line = line.split('=', 1)
                elif ':' in line:
                    # bar 0: i/o at 0xc280 [0xc2bf]
                    line = line.split(':', 1)
                elif ' ' in line:
                    # mmio ffffffffffffffff/0000000000100000
                    line = line.split(' ', 1)
                    # HOOK: mmio can have multiple values
                    if line[0] == 'mmio':
                        if 'mmio' not in current.qtree:
                            current.set_qtree_prop('mmio', [])
                        current.qtree['mmio'].append(line[1])
                        line = None
                    # old version: gpio-in 24, gpio-out 1
                    # new version: gpio-in "" 1, gpio-out "a20" 1
                    elif line[0] == 'gpio-in' or line[0] == 'gpio-out':
                        if '"' in line[1]:
                            gpio_name, line[1] = line[1].split(' ', 1)
                            line[0] += " %s" % gpio_name
                else:
                    # Corrupted qtree
                    raise ValueError('qtree line not recognized:\n%s' % line)
                if line:
                    current.set_qtree_prop(line[0].strip(), line[1].strip())
                if len(info) == 0:
                    break
                line = info.pop(0)
            else:
                # Node can be of different type
                current = _replace_node(current, current.guess_type())
                self.nodes.append(current)
                current = current.get_parent()
                offset -= OFFSET_PER_LEVEL
        # Read out remaining self.nodes
        while offset > 0:
            current = _replace_node(current, current.guess_type())
            self.nodes.append(current)
            current = current.get_parent()
            offset -= OFFSET_PER_LEVEL
        # This is the place to put HOOKs for nasty qtree devices
        for i in xrange(len(self.nodes)):
            _hook_usb2_disk(self.nodes[i])


class QtreeDisksContainer(object):

    """
    Container for QtreeDisks verification.
    It's necessary because some information can be verified only from
    informations about all disks, not only from single disk.
    """

    def __init__(self, nodes):
        """ work only with QtreeDisks instances """
        self.disks = []
        self.errors = []
        for node in nodes:
            if isinstance(node, QtreeDisk):
                if node.get_qname() != '<null>':
                    self.disks.append(node)

    def parse_info_block(self, info):
        """
        Extracts all information about self.disks and fills them in.

        :param info: output of ``info block`` command
        :return: ``self.disks`` defined in qtree but not in ``info block``,
                 ``self.disks`` defined in ``block info`` but not in qtree
        """
        additional = 0
        missing = 0
        for i in xrange(len(self.disks)):
            disk = self.disks[i]
            name = disk.get_qname()
            if name not in info:
                error_msg = "disk %s is in block but not in qtree" % name
                logging.error(error_msg)
                self.errors.append(error_msg)
                missing += 1
                continue
            for prop, value in six.iteritems(info[name]):
                disk.set_block_prop(prop, value)
        for disk in self.disks:
            if disk.get_block() == {}:
                error_msg = "disk in qtree but not in info block\n%s" % disk
                logging.error(error_msg)
                self.errors.append(error_msg)
                additional += 1
        return (additional, missing)

    def generate_params(self):
        """
        Generate params from current self.qtree and self.block info.
        :note: disk name is not yet the one from autotest params
        :return: number of fails
        """
        err = 0
        for disk in self.disks:
            try:
                disk.generate_params()
            except ValueError:
                error_msg = "generate_params error: %s" % disk
                logging.error(error_msg)
                self.errors.append(error_msg)
                err += 1
        return err

    def check_guests_proc_scsi(self, info):
        """
        Check info from guest's /proc/scsi/scsi file with qtree/block info

        :note: Not tested disks are of different type (virtio_blk, ...)
        :param info: contents of guest's /proc/scsi/scsi file
        :return: Number of disks missing in guest os, disks missing in qtree,
                 disks not tested from qtree, disks not tested from guest)
        """
        # Check only channel, id and lun for now
        additional = 0
        missing = 0
        qtree_not_scsi = 0
        proc_not_scsi = 0
        # host, channel, id, lun, vendor
        _scsis = re.findall(r'Host:\s+(\w+)\s+Channel:\s+(\d+)\s+Id:\s+(\d+)'
                            '\s+Lun:\s+(\d+)\n\s+Vendor:\s+([a-zA-Z0-9_-]+)'
                            '\s+Model:.*\n.*Type:\s+([a-zA-Z0-9_-]+)', info)
        disks = set()
        # Check only scsi disks
        for disk in self.disks:
            if (disk.get_qtree()['type'].startswith('scsi') or
                    disk.get_qtree()['type'].startswith('usb2')):
                props = disk.get_qtree()
                # New output from qtree will include hex number. Should
                # remove it in this function.
                for item in props:
                    if re.match("\d+\s+\(.*?\)", props[item]):
                        props[item] = re.findall("\d+", props[item])[0]

                disks.add('%d-%d-%d' % (int(props.get('channel')),
                                        int(props.get('scsi-id')),
                                        int(props.get('lun'))))
            else:
                qtree_not_scsi += 1
        scsis = set()
        for scsi in _scsis:
            # Ignore IDE disks
            if scsi[5] != 'CD-ROM':
                # Qemu encode LUNs with scsi's with 'flat space addressing'
                # method if LUNs > 255, decode LUNs when LUN ID - 16384 > 255.
                lun_id = int(scsi[3])
                if lun_id > 255 + 16384:
                    lun_id -= 16384
                scsis.add("%d-%d-%d" % (int(scsi[1]), int(scsi[2]), lun_id))
            else:
                proc_not_scsi += 1
        for disk in disks.difference(scsis):
            error_msg = 'Disk %s is in qtree but not in /proc/scsi/scsi' % disk
            logging.error(error_msg)
            self.errors.append(error_msg)
            additional += 1
        for disk in scsis.difference(disks):
            error_msg = 'Disk %s is in /proc/scsi/scsi but not in qtree' % disk
            logging.error(error_msg)
            self.errors.append(error_msg)
            missing += 1
        return (additional, missing, qtree_not_scsi, proc_not_scsi)

    def check_disk_params(self, params):
        """
        Check gathered info from qtree/block with params
        :param params: autotest params
        :return: number of errors
        """
        def check_drive_format(node, params):
            """ checks the drive format according to qtree info """
            expected = params.get('drive_format')
            if expected == 'scsi':
                if arch.ARCH in ('ppc64', 'ppc64le'):
                    expected = 'spapr-vscsi'
                else:
                    expected = 'lsi53c895a'
            elif expected.startswith('scsi'):
                expected = params.get('scsi_hba', 'virtio-scsi-pci')
            elif expected.startswith('usb'):
                expected = 'usb-storage'
            try:
                if expected == 'virtio':
                    actual = node.qtree['type']
                else:
                    actual = node.parent.parent.qtree.get('type')
            except AttributeError:
                error_msg = "Failed to check drive format,"
                error_msg += " can't get parent of:\n%s" % node
                logging.error(error_msg)
                self.errors.append(error_msg)
            if actual == 'virtio-scsi-device':  # new name for virtio-scsi
                actual = 'virtio-scsi-pci'
            if expected not in actual:
                return ("drive format in qemu is %s, in autotest %s"
                        % (actual, expected))

        err = 0
        disks = {}
        for disk in self.disks:
            if isinstance(disk, QtreeDisk):
                disks[disk.get_qname()] = (disk.get_params().copy(), disk)
        # We don't have the params name so we need to map file_names instead
        qname = None
        for name in params.objects('cdroms'):
            image_name = utils_misc.get_path(
                data_dir.get_data_dir(),
                params.object_params(name).get('cdrom', ''))
            image_name = os.path.realpath(image_name)
            for (qname, disk) in six.iteritems(disks):
                if disk[0].get('image_name') == image_name:
                    break
            else:
                continue    # Not /proc/scsi cdrom device
            disks.pop(qname)
        for name in params.objects('images'):
            current = None
            image_params = params.object_params(name)

            base_dir = image_params.get("images_base_dir",
                                        data_dir.get_data_dir())

            image_name = os.path.realpath(
                storage.get_image_filename(image_params,
                                           base_dir))
            for (qname, disk) in six.iteritems(disks):
                if disk[0].get('image_name') == image_name:
                    current = disk[0]
                    current_node = disk[1]
                    # autotest params might use relative path
                    current['image_name'] = image_params.get('image_name')
                    break
            if not current:
                error_msg = "Disk %s is not in qtree but is in params." % name
                logging.error(error_msg)
                self.errors.append(error_msg)
                err += 1
                continue
            for prop in six.iterkeys(current):
                handled = False
                if prop == "drive_format":
                    out = check_drive_format(current_node, image_params)
                    if out:
                        error_msg = "Disk %s %s" % (qname, out)
                        logging.error(error_msg)
                        self.errors.append(error_msg)
                        err += 1
                    handled = True
                elif (image_params.get(prop) and
                        image_params.get(prop) == current.get(prop)):
                    handled = True
                if not handled:
                    error_msg = "Disk %s property " % qname
                    error_msg += "%s=%s doesn't match" % (
                        prop, current.get(prop))
                    error_msg += " params %s" % image_params.get(prop)
                    logging.error(error_msg)
                    self.errors.append(error_msg)
                    err += 1
            disks.pop(qname)
        if disks:
            error_msg = "Some disks were in qtree but not in"
            error_msg += " autotest params: %s" % disks
            logging.error(error_msg)
            self.errors.append(error_msg)
            err += 1
        return err


def traverse(qtree_node):
    """
    Use generator to traverse the given qtree node

    :param qtree_node: qtree node to be traversed
    """
    if not qtree_node:
        return

    # Traverse in pre-order, please fix this
    # if there are some special demands
    yield qtree_node
    for sub_node in qtree_node.get_children():
        for node in traverse(sub_node):
            yield node



"""
Shared code for tests that need to get the libvirt version
"""

import re
import logging

from avocado.utils import path
from avocado.utils import process
from avocado.utils.astring import to_text


def version_compare(major, minor, update, session=None):
    """
    Determine/use the current libvirt library version on the system
    and compare input major, minor, and update values against it.
    If the running version is greater than or equal to the input
    params version, then return True; otherwise, return False

    This is designed to handle upstream version comparisons for
    test adjustments and/or comparisons as a result of upstream
    fixes or changes that could impact test results.

    :param major: Major version to compare against
    :param minor: Minor version to compare against
    :param update: Update value to compare against
    :param session: Shell session on remote host
    :return: True if running version is greater than or
                  equal to the input libvirt version
    """
    LIBVIRT_LIB_VERSION = 0

    func = process.system_output
    if session:
        func = session.cmd_output
        cmd = "virtqemud"
        if session.cmd_status('which %s' % cmd):
            cmd = "libvirtd"
    else:
        try:
            path.find_command("virtqemud")
            cmd = "virtqemud"
        except path.CmdNotFoundError:
            cmd = "libvirtd"

    try:
        regex = r'\w*d\s*\(libvirt\)\s*'
        regex += r'(\d+)\.(\d+)\.(\d+)'
        lines = to_text(func("%s -V" % cmd)).splitlines()
        for line in lines:
            mobj = re.search(regex, line, re.I)
            if bool(mobj):
                LIBVIRT_LIB_VERSION = int(mobj.group(1)) * 1000000 + \
                                      int(mobj.group(2)) * 1000 + \
                                      int(mobj.group(3))
                break
    except (ValueError, TypeError, AttributeError):
        logging.warning("Error determining libvirt version")
        return False

    compare_version = major * 1000000 + minor * 1000 + update
    if LIBVIRT_LIB_VERSION == 0:
        logging.error("Unable to get virtqemud/libvirtd version!")
    elif LIBVIRT_LIB_VERSION >= compare_version:
        return True
    return False
